from __future__ import annotations

import logging
import textwrap
from pathlib import Path

from ...commands.cmake import CMaker
from ...common import Config, Module
from ...common.util import get_os_name

logger = logging.getLogger(__name__)


def write_build_hook(cfg: Config, staging_dir: Path, module: Module, cmaker: CMaker):
    """Write a hook that re-compiles extension modules."""
    edit_cfg = cfg.editable["cross" if cfg.cross else get_os_name()]
    if not edit_cfg.get("build_hook"):
        return
    if edit_cfg.get("mode") != "symlink":
        logger.warning("Skipping build_hook: only supported for symlink mode")
        return
    name = module.name
    pkg_hook = staging_dir / (name + "_build_hook")
    pkg_hook.mkdir(parents=True, exist_ok=True)
    cwd = cmaker.get_working_dir()
    env = cmaker.conf_settings.environment or {}
    cmd = list(cmaker.get_build_commands()) + list(cmaker.get_install_commands())
    content = f"""\
        import sys, inspect, os
        from importlib.machinery import PathFinder
        from string import Template
        import subprocess

        class BuilderPathFinder(PathFinder):
            def __init__(self, name, cwd, env, cmd):
                self.name = name
                self.cwd = cwd
                self.env = env
                self.cmd = cmd
            def find_spec(self, name, path=None, target=None):
                if name.split('.', 1)[0] == self.name:
                    self.build()
                return None
            def prepare_environment(self):
                env = os.environ.copy()
                for k, v in self.env.items():
                    templ = Template(v)
                    env[k] = templ.substitute(env)
                return env
            def build(self):
                env = self.prepare_environment()
                for cmd in self.cmd:
                    try:
                        subprocess.run(cmd, cwd=self.cwd, check=True, env=env)
                    except subprocess.CalledProcessError as e:
                        raise ImportError(
                            f"Failed to build dependencies for module {{self.name!r}}",
                            name=self.name,
                            path=self.cwd,
                        ) from e
                sys.meta_path.remove(self)

        def install(name: str, cwd, env, cmd):
            sys.meta_path.insert(0, BuilderPathFinder(name, cwd, env, cmd))

        install(
            name={name!r},
            cwd={cwd!r},
            env={env!r},
            cmd={cmd!r},
        )
        """
    (pkg_hook / "__init__.py").write_text(textwrap.dedent(content), encoding="utf-8")
    # Write a path file to find the development files
    content = f"import {name}_build_hook\n"
    with (staging_dir / f"{name}.pth").open("a") as f:
        f.write(content)



from __future__ import annotations

import re
import sys
from pprint import pprint
from subprocess import CalledProcessError
from subprocess import run as sp_run

from distlib.version import NormalizedVersion  # type: ignore[import-untyped]


class CommandRunner:
    def __init__(self, verbose: bool = False, dry: bool = False):
        self.verbose = verbose
        self.dry = dry

    def run(self, *args, **kwargs):
        """Wrapper around subprocess.run that optionally prints the command."""
        if self.verbose:
            pprint([*args])
            pprint(kwargs)
            print(flush=True)
        elif self.dry:
            from shlex import join

            print(join(args[0]))
        if not self.dry:
            return sp_run(*args, **kwargs)  # noqa: PLW1510
        return None

    def check_program_version(
        self,
        program: str,
        minimum_version: NormalizedVersion | None,
        name: str | None,
        check_version: bool = True,
    ):
        """Check if there's a new enough version of the given command available
        in PATH."""
        name = name or program
        try:
            # Try running the command
            cmd = [program, "--version"] if check_version else [program, "-h"]
            res = self.run(cmd, check=True, capture_output=True, encoding="utf-8")
            # Try finding the version
            if res is not None and check_version:
                m = re.search(r"\d+(\.\d+){1,}", res.stdout)
                if not m:
                    msg = f"Unexpected {name} version output"
                    raise RuntimeError(msg)
                program_version = NormalizedVersion(m.group(0))
                if self.verbose:
                    print("Found", name, program_version)
                # Check if the version is new enough
                if minimum_version is not None and program_version < minimum_version:
                    msg = f"{name} too old"
                    raise RuntimeError(msg)
        except CalledProcessError as e:
            if self.verbose:
                print(f"{type(e).__module__}.{type(e).__name__}", e, sep=": ")
                print(e.stdout)
                print(e.stderr, file=sys.stderr)
            return False
        except Exception as e:
            # If any of that failed, return False
            if self.verbose:
                print(f"{type(e).__module__}.{type(e).__name__}", e, sep=": ")
            return False
        return True



from __future__ import annotations

from typing import Any

from .config_path import ConfPath


class ValueReference:
    def __init__(self, value_path: ConfPath, values: dict) -> None:
        self.value_path = value_path
        self.values: dict | Any = values

    def is_value_set(self, path: str | ConfPath):
        if isinstance(path, str):
            return path in self.values
        values = self.values
        while path:
            name, path = path.split_front()
            if name not in values:
                return False
            values = values[name]
        return True

    def get_value(self, path: str | ConfPath):
        if isinstance(path, str):
            return self.values[path]
        values = self.values
        while path:
            name, path = path.split_front()
            values = values[name]
        return values

    def set_value(self, path: str | ConfPath, val: Any):
        if isinstance(path, str):
            self.values[path] = val
            return True
        values = self.values
        while True:
            name, path = path.split_front()
            if not path:
                values[name] = val
                return True
            if name not in values:
                return False
            values = values[name]

    def set_value_default(self, path: str | ConfPath, val: Any):
        if isinstance(path, str):
            self.values.setdefault(path, val)
            return True
        values = self.values
        while True:
            name, path = path.split_front()
            if not path:
                values.setdefault(name, val)
                return True
            if name not in values:
                return False
            values = values[name]

    def sub_ref(self, name: str | ConfPath) -> ValueReference:
        if isinstance(name, ConfPath) and len(name.pth) == 1:
            name = name.pth[0]
        if isinstance(name, str):
            if not self.is_value_set(name):
                raise KeyError(name)
            return ValueReference(
                value_path=self.value_path.join(name),
                values=self.values[name],
            )
        name, rem = name.split_front()
        return self.sub_ref(name).sub_ref(rem)



from __future__ import annotations

import logging
import shutil
from copy import copy
from pathlib import Path

from ..common import Config

logger = logging.getLogger(__name__)


def _write_entry_points(entrypoints: dict[str, dict[str, str]], fp):
    """Write entry_points.txt from a two-level dict

    Sorts on keys to ensure results are reproducible.

    Based on flit_core.
    """
    for group_name in sorted(entrypoints):
        fp.write(f"[{group_name}]\n")
        group = entrypoints[group_name]
        for name in sorted(group):
            val = group[name]
            fp.write(f"{name}={val}\n")
        fp.write("\n")


def write_entry_points(cfg: Config, distinfo_dir: Path):
    entrypoints = copy(cfg.standard_metadata.entrypoints)
    if cfg.standard_metadata.scripts:
        entrypoints["console_scripts"] = cfg.standard_metadata.scripts
    if cfg.standard_metadata.gui_scripts:
        entrypoints["gui_scripts"] = cfg.standard_metadata.gui_scripts
    with (distinfo_dir / "entry_points.txt").open("w", encoding="utf-8") as f:
        _write_entry_points(entrypoints, f)


def write_license_files(cfg: Config, distinfo_dir: Path):
    """Write the LICENSE file from pyproject.toml to the distinfo
    directory."""
    license = cfg.standard_metadata.license
    if not license:
        return
    if license.file:
        shutil.copy2(license.file, distinfo_dir)
    else:
        with (distinfo_dir / "LICENSE").open("w", encoding="utf-8") as f:
            f.write(license.text)


def write_metadata(cfg: Config, distinfo_dir: Path):
    metadata_path = distinfo_dir / "METADATA"
    with metadata_path.open("w", encoding="utf-8") as f:
        f.write(str(cfg.standard_metadata.as_rfc822()))



from __future__ import annotations

import textwrap
from pathlib import Path

from ...common import Module


def write_editable_hook(staging_dir: Path, module: Module):
    """Write a hook that finds the installed compiled extension modules."""
    name = module.name
    pkg_hook = staging_dir / (name + "_editable_hook")
    pkg_hook.mkdir(parents=True, exist_ok=True)
    content = f"""\
        import sys, inspect, os
        from importlib.machinery import PathFinder

        class EditablePathFinder(PathFinder):
            def __init__(self, name, extra_path):
                self.name = name
                self.extra_path = extra_path
            def find_spec(self, name, path=None, target=None):
                if name.split('.', 1)[0] != self.name:
                    return None
                if path is None:
                    path = []
                path.append(self.extra_path)
                return super().find_spec(name, path, target)

        def install(name: str):
            source_path = os.path.abspath(inspect.getsourcefile(EditablePathFinder))
            source_dir = os.path.dirname(source_path)
            installed_path = os.path.join(source_dir, '..', name)
            sys.meta_path.insert(0, EditablePathFinder(name, installed_path))

        install('{name}')
        """
    (pkg_hook / "__init__.py").write_text(textwrap.dedent(content), encoding="utf-8")
    # Write a path file to find the development files
    content = f"{module.full_path.parent!s}\n" f"import {name}_editable_hook\n"
    (staging_dir / f"{name}.pth").write_text(textwrap.dedent(content))



from __future__ import annotations

import logging
import os
import shutil
import tempfile
from pathlib import Path
from typing import Any

from distlib.wheel import Wheel  # type: ignore[import-untyped]

from .commands.cmake import (
    CMakeBuildSettings,
    CMakeConfigureSettings,
    CMakeInstallSettings,
    CMaker,
    CMakeSettings,
)
from .commands.cmd_runner import CommandRunner
from .commands.try_run import check_cmake_program, check_stubgen_program
from .common import (
    BuildPaths,
    ComponentConfig,
    Config,
    ConfigError,
    FormattedErrorMessage,
    Module,
    PackageInfo,
    ProblemInModule,
    format_and_rethrow_exception,
    logformat,
    util,
)
from .config import load as config_load
from .config.dynamic import find_module, update_dynamic_metadata
from .export import editable as export_editable
from .export import metadata as export_metadata
from .export import util as export_util
from .export.editable.build_hook import write_build_hook
from .export.native_tags import get_interpreter_name
from .export.sdist import SdistBuilder
from .export.tags import convert_wheel_tags, get_cross_tags, get_native_tags, is_pure

logger = logging.getLogger(__name__)


class _BuildBackend:
    # --- Constructor ---------------------------------------------------------

    def __init__(self) -> None:
        self.runner: CommandRunner = CommandRunner()

    @property
    def verbose(self):
        return self.runner.verbose

    # --- Methods required by PEP 517 -----------------------------------------

    def get_requires_for_build_wheel(self, config_settings=None):
        """https://www.python.org/dev/peps/pep-0517/#get-requires-for-build-wheel"""
        try:
            self.parse_config_settings(config_settings)

            src_dir = Path().resolve()
            cfg = self.read_config(src_dir, config_settings, self.verbose)
            return self.get_requires_build_project(config_settings, cfg, self.runner)
        except Exception as e:
            format_and_rethrow_exception(e)

    def get_requires_for_build_editable(self, config_settings=None):
        """https://www.python.org/dev/peps/pep-0660/#get-requires-for-build-editable"""
        return self.get_requires_for_build_wheel(config_settings)

    def get_requires_for_build_sdist(self, config_settings=None):
        """https://www.python.org/dev/peps/pep-0517/#get-requires-for-build-sdist"""
        return []

    def build_wheel(
        self, wheel_directory, config_settings=None, metadata_directory=None
    ):
        """https://www.python.org/dev/peps/pep-0517/#build-wheel"""
        try:
            assert metadata_directory is None

            # Parse options
            self.parse_config_settings(config_settings)

            # Build wheel
            with tempfile.TemporaryDirectory() as tmp_build_dir:
                return self.build_wheel_in_dir(
                    wheel_directory, tmp_build_dir, config_settings
                )
        except Exception as e:
            format_and_rethrow_exception(e)

    def build_editable(
        self, wheel_directory, config_settings=None, metadata_directory=None
    ):
        """https://www.python.org/dev/peps/pep-0660/#build-editable"""
        try:
            assert metadata_directory is None

            # Parse options
            self.parse_config_settings(config_settings)

            # Build wheel
            with tempfile.TemporaryDirectory() as tmp_build_dir:
                return self.build_wheel_in_dir(
                    wheel_directory, tmp_build_dir, config_settings, editable=True
                )
        except Exception as e:
            format_and_rethrow_exception(e)

    def build_sdist(self, sdist_directory, config_settings=None):
        """https://www.python.org/dev/peps/pep-0517/#build-sdist"""
        try:
            # Parse options
            self.parse_config_settings(config_settings)

            return self.do_build_sdist(sdist_directory, config_settings)
        except Exception as e:
            format_and_rethrow_exception(e)

    # --- Parsing config options and metadata ---------------------------------

    @staticmethod
    def is_verbose_enabled(config_settings: dict | None):
        truthy = lambda x: x.lower() in ("", "1", "true", "yes", "y")
        if config_settings is not None:
            verbose_keys = {"verbose", "--verbose", "V", "-V"}
            verbose_opts = {
                k: v for k, v in config_settings.items() if k in verbose_keys
            }
            if verbose_opts:
                last_val = next(reversed(list(verbose_opts.values())))
                return truthy(last_val)
        env_verbose = os.environ.get("PY_BUILD_CMAKE_VERBOSE")
        if env_verbose is not None:
            return truthy(env_verbose)
        return False

    @staticmethod
    def get_log_level(config_settings: dict | None) -> int:
        def parse_log_level(loglevel: str) -> int:
            numeric_level = getattr(logging, loglevel.upper(), None)
            if isinstance(numeric_level, int):
                return numeric_level
            raise ValueError("Invalid log level: %s" % loglevel)

        if config_settings is not None:
            log_keys = {"loglevel", "--loglevel"}
            log_opts = {k: v for k, v in config_settings.items() if k in log_keys}
            if log_opts:
                last_val = next(reversed(list(log_opts.values())))
                return parse_log_level(last_val)
        env_log = os.environ.get("PY_BUILD_CMAKE_LOGLEVEL")
        if env_log is not None:
            return parse_log_level(env_log)
        return logging.INFO

    def parse_config_settings(self, config_settings: dict | None):
        try:
            level = self.get_log_level(config_settings)
            if "GITHUB_ACTIONS" in os.environ:
                formatter = logformat.GitHubActionsFormatter()
                handler = logging.StreamHandler()
                handler.setFormatter(formatter)
                logging.basicConfig(level=level, handlers=[handler])
            else:
                logging.basicConfig(level=level)
        except ValueError as e:
            logger.error("Invalid log level specified", exc_info=e)
        self.runner.verbose = self.is_verbose_enabled(config_settings)

    @staticmethod
    def get_requires_build_project(
        config_settings: dict | None, cfg: Config, runner: CommandRunner
    ):
        deps: list[str] = []
        # Check if we need CMake
        if cfg.cmake:
            check_cmake_program(cfg, deps, runner)
        if cfg.stubgen:
            check_stubgen_program(deps, runner)
        if runner.verbose:
            print("Dependencies for build:", deps)
        return deps

    # --- Building wheels -----------------------------------------------------

    def build_wheel_in_dir(
        self, wheel_dir, tmp_build_dir, config_settings, editable=False
    ):
        """This is the main function that contains all steps necessary to build
        a complete wheel package, including the CMake builds etc."""

        # Load metadata from the pyproject.toml file
        src_dir = Path().resolve()
        cfg, module = self.read_all_metadata(src_dir, config_settings, self.verbose)
        pkg_info = self.get_pkg_info(cfg, module)
        cmake_cfg = self.get_cmake_config(cfg)

        # Set up all paths
        paths = self.get_default_paths(
            wheel_dir, tmp_build_dir, src_dir, cfg, cmake_cfg
        )

        # Copy the module's Python source files to the temporary folder
        if not editable:
            export_util.copy_pkg_source_to(paths.staging_dir, module)
        else:
            paths = export_editable.do_editable_install(cfg, paths, module)

        # Create dist-info folder
        distinfo_dir = f"{pkg_info.norm_name}-{pkg_info.version}.dist-info"
        distinfo_dir = paths.pkg_staging_dir / distinfo_dir
        distinfo_dir.mkdir(parents=True, exist_ok=True)

        # Write metadata, license and entry points to Wheel's distinfo
        export_metadata.write_metadata(cfg, distinfo_dir)
        export_metadata.write_license_files(cfg, distinfo_dir)
        export_metadata.write_entry_points(cfg, distinfo_dir)

        # Configure, build and install the CMake project
        if cmake_cfg:
            cmaker = self.get_cmaker(
                paths.source_dir,
                paths.build_dir,
                paths.staging_dir,
                cmake_cfg,
                cfg.cross,
                pkg_info,
                runner=self.runner,
            )
            cmaker.configure()
            cmaker.build()
            cmaker.install()

            if editable:
                write_build_hook(cfg, paths.pkg_staging_dir, module, cmaker)

        # Generate .pyi stubs (for the Python files only)
        if cfg.stubgen is not None and not editable:
            self.generate_stubs(paths, module, cfg.stubgen)

        # Create wheel
        return self.create_wheel(paths, cfg, cmake_cfg, pkg_info)

    @staticmethod
    def get_pkg_info(cfg: Config | ComponentConfig, module: Module | None):
        return PackageInfo(
            version=str(cfg.standard_metadata.version),
            package_name=cfg.package_name,
            module_name=module.name if module is not None else "",
        )

    @staticmethod
    def get_default_paths(wheel_dir, tmp_build_dir, src_dir, cfg, cmake_cfg):
        build_cfg_name = _BuildBackend.get_build_config_name(cfg.cross)
        if cmake_cfg:
            path = cmake_cfg["build_path"]
            build_dir = Path(str(path).replace("{build_config}", build_cfg_name))
        else:
            build_dir = src_dir / ".py-build-cmake_cache" / build_cfg_name
        return BuildPaths(
            source_dir=src_dir,
            build_dir=build_dir,
            wheel_dir=Path(wheel_dir),
            temp_dir=Path(tmp_build_dir),
            staging_dir=Path(tmp_build_dir) / "staging",
            pkg_staging_dir=Path(tmp_build_dir) / "staging",
        )

    @staticmethod
    def read_all_metadata(src_dir, config_settings, verbose) -> tuple[Config, Module]:
        cfg = _BuildBackend.read_config(src_dir, config_settings, verbose)
        module = find_module(cfg.module, src_dir)
        modfile = module.full_file
        if cfg.standard_metadata.dynamic:
            if module.is_generated:
                msg = "Dynamic metadata is not supported for generated modules/packages"
                raise ConfigError(msg)
            elif module.is_namespace:
                msg = "Dynamic metadata is not supported for namespace packages"
                raise ConfigError(msg)
        try:
            update_dynamic_metadata(cfg.standard_metadata, modfile)
        except ImportError as e:
            logger.error("Error importing %s for reading metadata", str(modfile))
            msg = (
                "\n"
                "\n"
                f"\t\u274C Error importing {modfile} for reading metadata:\n"
                "\n"
                f"\t\t{e.msg}\n"
                "\n"
            )
            raise FormattedErrorMessage(msg) from e
        except ProblemInModule as e:
            logger.error("Error reading metadata from %s", str(modfile))
            msg = (
                "\n"
                "\n"
                f"\t\u274C Error reading metadata from {modfile}:"
                "\n"
                "\n"
                f"\t\t{e}\n"
                "\n"
            )
            raise FormattedErrorMessage(msg) from e
        return cfg, module

    @staticmethod
    def read_config(src_dir, config_settings, verbose):
        """Read the configuration without the dynamic data."""
        return config_load.read_full_config(
            src_dir / "pyproject.toml", config_settings, verbose
        )

    @staticmethod
    def create_wheel(
        paths: BuildPaths, cfg: Config, cmake_cfg, package_info: PackageInfo
    ):
        """Create a wheel package from the build directory."""
        whl = Wheel()
        whl.name = package_info.norm_name
        whl.version = package_info.version
        pure = is_pure(cmake_cfg)
        libdir = "purelib" if pure else "platlib"
        staging_dir = paths.pkg_staging_dir
        whl_paths = {"prefix": str(staging_dir), libdir: str(staging_dir)}
        whl.dirname = paths.wheel_dir
        if pure:
            tags = {"pyver": ["py3"]}
        elif cfg.cross:
            tags = get_cross_tags(cfg.cross)
            tags = convert_wheel_tags(tags, cmake_cfg)
        else:
            tags = get_native_tags()
            tags = convert_wheel_tags(tags, cmake_cfg)
        wheel_path = whl.build(whl_paths, tags=tags, wheel_version=(1, 0))
        logger.debug("Built Wheel: %s", wheel_path)
        return str(Path(wheel_path).relative_to(paths.wheel_dir))

    @staticmethod
    def get_cmake_config(cfg: Config):
        if not cfg.cmake:
            return None
        if cfg.cross is None:
            return cfg.cmake[util.get_os_name()]
        else:
            return cfg.cmake["cross"]

    # --- Building sdists -----------------------------------------------------

    def do_build_sdist(self, sdist_directory, config_settings):
        # Load metadata
        src_dir = Path().resolve()
        pyproject = src_dir / "pyproject.toml"
        cfg, module = self.read_all_metadata(src_dir, config_settings, self.verbose)
        pkg_info = self.get_pkg_info(cfg, module)

        # Export dist
        extra_files = [pyproject, *cfg.referenced_files]
        sdist_cfg = cfg.sdist["cross" if cfg.cross else util.get_os_name()]
        sdist_builder = SdistBuilder(
            module,
            pkg_info,
            metadata=cfg.standard_metadata,
            cfgdir=src_dir,
            extra_files=extra_files,
            include_patterns=sdist_cfg.get("include_patterns", []),
            exclude_patterns=sdist_cfg.get("exclude_patterns", []),
        )
        sdist_tar = sdist_builder.build(Path(sdist_directory))
        return str(Path(sdist_tar).relative_to(sdist_directory))

    # --- CMake builds --------------------------------------------------------

    @staticmethod
    def get_cmaker(
        source_dir: Path,
        build_dir: Path,
        install_dir: Path | None,
        cmake_cfg: dict,
        cross_cfg: dict | None,
        package_info: PackageInfo,
        **kwargs,
    ):
        # Optionally include the cross-compilation settings
        if cross_cfg:
            cross_compiling = True

            def cvt_path(x):
                if x is None:
                    return None
                if isinstance(x, str) and not x:
                    return None
                assert isinstance(x, (Path, str))
                return Path(x)

            def python_tag_to_cmake(x: str | None):
                if x is None:  # If not set, use native interpreter
                    x = get_interpreter_name()
                return {
                    "cp": "Python",
                    "pp": "PyPy",
                    "ip": "IronPython",
                    "jy": "Jython",
                }.get(x)

            cross_opts = {
                "toolchain_file": cvt_path(cross_cfg.get("toolchain_file")),
                "python_prefix": cvt_path(cross_cfg.get("prefix")),
                "python_library": cvt_path(cross_cfg.get("library")),
                "python_include_dir": cvt_path(cross_cfg.get("include_dir")),
                "python_interpreter_id": python_tag_to_cmake(
                    cross_cfg.get("implementation")
                ),
            }
        else:
            cross_compiling = False
            cross_opts = {
                "toolchain_file": None,
                "python_prefix": None,
                "python_library": None,
                "python_include_dir": None,
                "python_interpreter_id": None,
            }

        # Add some CMake configure options
        options = cmake_cfg.get("options", {})
        btype = cmake_cfg.get("build_type")
        if btype:  # -D CMAKE_BUILD_TYPE={type}
            options["CMAKE_BUILD_TYPE:STRING"] = btype

        # CMake options
        return CMaker(
            cmake_settings=CMakeSettings(
                working_dir=source_dir,
                source_path=Path(cmake_cfg["source_path"]),
                build_path=build_dir,
                os=util.get_os_name(),
                find_python=bool(cmake_cfg["find_python"]),
                find_python3=bool(cmake_cfg["find_python3"]),
                minimum_required=cmake_cfg["minimum_version"],
                command=Path("cmake"),
            ),
            conf_settings=CMakeConfigureSettings(
                environment=cmake_cfg.get("env", {}),
                build_type=cmake_cfg.get("build_type"),
                options=options,
                args=cmake_cfg.get("args", []),
                preset=cmake_cfg.get("preset"),
                generator=cmake_cfg.get("generator"),
                cross_compiling=cross_compiling,
                **cross_opts,
            ),
            build_settings=CMakeBuildSettings(
                args=cmake_cfg["build_args"],
                tool_args=cmake_cfg["build_tool_args"],
                presets=cmake_cfg.get("build_presets", []),
                configs=cmake_cfg.get("config", []),
            ),
            install_settings=CMakeInstallSettings(
                args=cmake_cfg["install_args"],
                configs=cmake_cfg.get("config", []),
                components=cmake_cfg.get("install_components", []),
                prefix=install_dir,
            ),
            package_info=package_info,
            **kwargs,
        )

    # --- Generate stubs ------------------------------------------------------

    def generate_stubs(self, paths: BuildPaths, module: Module, cfg: dict[str, Any]):
        """Generate stubs (.pyi) using mypy stubgen."""
        stubs_dir = paths.temp_dir / "stubs"
        args = ["stubgen", *cfg.get("args", [])]
        is_package = module.is_package
        cfg.setdefault("packages", [module.name] if is_package else [])
        for p in cfg["packages"]:
            args += ["-p", p]
        cfg.setdefault("modules", [module.name] if not is_package else [])
        for m in cfg["modules"]:
            args += ["-m", m]
        args += cfg.get("files", [])
        # Add output folder argument if not already specified in cfg['args']
        if "args" not in cfg or not ({"-o", "--output"} & set(cfg["args"])):
            args += ["-o", str(stubs_dir)]
        # Add search path argument if not already specified in cfg['args']
        if "args" not in cfg or "--search-path" not in cfg["args"]:
            args += ["--search-path", str(paths.staging_dir)]
        env = os.environ.copy()
        env.setdefault("MYPY_CACHE_DIR", str(paths.temp_dir))
        # Call mypy stubgen in a subprocess
        self.runner.run(args, cwd=paths.staging_dir, check=True, env=env)
        # Copy the stubs to the staging folder (careful not to overwrite
        # existing stubs generated by CMake)
        self.copy_stubs(stubs_dir, paths)

    def copy_stubs(self, stubs_dir: Path, paths: BuildPaths):
        dest = paths.staging_dir

        def stubs_already_exists(dest_path: Path):
            # We don't want to replace existing .pyi files.
            dest_pyi = dest_path.with_suffix(".pyi")
            if dest_pyi.exists():
                return True
            # If a directory with the same name already exists, we only
            # want to copy our .pyi file if the directory does not contain
            # an __init__.pyi file.
            dest_dir = dest_path.with_suffix("")
            dest_dir_init = dest_dir / "__init__.pyi"
            return dest_dir.is_dir() and dest_dir_init.exists()

        def handle_file(root_path: Path, f: str):
            rel_path = root_path.relative_to(stubs_dir)
            src_path = root_path / f
            dest_path = dest / rel_path / f
            if not stubs_already_exists(dest_path):
                logger.debug("Copying generated stub  %s -> %s", src_path, dest_path)
                shutil.move(str(src_path), str(dest_path))
            else:
                logger.info(
                    "Not copying generated stub file %s because a .pyi "
                    "file for the same module already exists",
                    rel_path / f,
                )

        def handle_dir(root_path: Path, d: str) -> bool:
            rel_path = root_path.relative_to(stubs_dir)
            src_path = root_path / d
            dest_path = dest / rel_path / d
            dest_pyi = dest_path.with_suffix(".pyi")
            # If the destination already has stubs for this (sub)module
            # in a .pyi file, adding the folder as well would only cause
            # confusion. Ignore the new folder and keep the existing
            # .pyi file. Don't recurse into the folder either.
            if dest_pyi.exists():
                logger.info(
                    "Not copying generated stub directory %s because a "
                    ".pyi file for the same module already exists",
                    rel_path / d,
                )
                return False
            # If there's already a folder with the same name, simply recurse
            # into it.
            if dest_path.exists():
                if dest_path.is_dir():
                    return True
                logger.debug(
                    "Not copying generated stub directory %s because a "
                    "file with the same name already exists",
                    rel_path / d,
                )
                return False
            # If there's neither a .pyi file nor a folder with the same
            # name, we can safely move our folder to the destination, and
            # there's no need to recurse any further.
            logger.debug("Copying generated stubs %s -> %s", src_path, dest_path)
            shutil.move(str(src_path), str(dest_path))
            return False

        for root, dirs, files in os.walk(stubs_dir):
            root_path = Path(root)
            is_pyi = lambda f: Path(f).suffix == ".pyi"
            for f in filter(is_pyi, files):
                handle_file(root_path, f)
            dirs[:] = [d for d in dirs if handle_dir(root_path, d)]

    # --- Misc helper functions -----------------------------------------------

    @staticmethod
    def get_build_config_name(cross_cfg):
        """Get a string representing the Python version, ABI and architecture,
        used to name the build folder so builds for different versions don't
        interfere."""
        tags = get_cross_tags(cross_cfg) if cross_cfg else get_native_tags()
        return "-".join(x[0] for x in tags.values())


_BACKEND = _BuildBackend()
get_requires_for_build_wheel = _BACKEND.get_requires_for_build_wheel
get_requires_for_build_sdist = _BACKEND.get_requires_for_build_sdist
get_requires_for_build_editable = _BACKEND.get_requires_for_build_editable
build_wheel = _BACKEND.build_wheel
build_sdist = _BACKEND.build_sdist
build_editable = _BACKEND.build_editable



from __future__ import annotations

from distlib.version import NormalizedVersion  # type: ignore[import-untyped]

from ..common import CMAKE_MINIMUM_REQUIRED, Config
from ..common.util import get_os_name
from .cmd_runner import CommandRunner


def check_cmake_program(cfg: Config, deps: list[str], runner: CommandRunner):
    assert cfg.cmake
    # Do we need to perform a native build?
    native = not cfg.cross
    native_cfg = cfg.cmake[get_os_name()] if native else {}
    # Do we need to perform a cross build?
    cross = cfg.cross
    cross_cfg = cfg.cmake.get("cross", {})
    # Find the strictest version requirement
    min_cmake_ver = max(
        NormalizedVersion(CMAKE_MINIMUM_REQUIRED),
        NormalizedVersion(native_cfg.get("minimum_version", "0.0")),
        NormalizedVersion(cross_cfg.get("minimum_version", "0.0")),
    )
    # If CMake in PATH doesn't work or is too old, add it as a build
    # requirement
    if not runner.check_program_version("cmake", min_cmake_ver, "CMake"):
        deps.append("cmake>=" + str(min_cmake_ver))

    # Check if we need Ninja
    cfgs = []
    if native:
        cfgs.append(native_cfg)
    if cross:
        cfgs.append(cross_cfg)

    # Do any of the configs require Ninja as a generator?
    needs_ninja = lambda c: "ninja" in c.get("generator", "").lower()
    need_ninja = any(map(needs_ninja, cfgs))
    if need_ninja and not runner.check_program_version("ninja", None, "Ninja"):
        # If so, check if a working version exists in the PATH, otherwise,
        # add it as a build requirement
        deps.append("ninja")


def check_stubgen_program(deps: list[str], runner: CommandRunner):
    if not runner.check_program_version("stubgen", None, None, False):
        # we need https://github.com/python/mypy/pull/14722
        deps.append("mypy>=1.4.0")



from __future__ import annotations

import logging
import os
from dataclasses import dataclass, field
from pathlib import Path
from subprocess import CalledProcessError
from typing import Any, Sequence

import pyproject_metadata

logger = logging.getLogger(__name__)

CMAKE_MINIMUM_REQUIRED = "3.15"


class FormattedErrorMessage(Exception):
    """Wrapper exception for any error that already has a nicely formatted
    error message."""


class ConfigError(ValueError):
    """Problem processing the pyproject.toml config"""


class ProblemInModule(ValueError):
    """Problem processing the project's modules"""


class NoDocstringError(ProblemInModule):
    """The module does not contain a docstring that can be used as package
    description."""


class NoVersionError(ProblemInModule):
    """The module does not contain a __version__."""


class InvalidVersion(ProblemInModule):
    """The __version__ is invalid."""


@dataclass
class Module:
    """Describes the location and name of a Python package or module"""

    name: str
    full_path: Path
    base_path: Path
    is_package: bool
    is_namespace: bool
    is_generated: bool = False

    @property
    def prefix(self):
        return self.full_path.parent

    @property
    def full_file(self):
        if self.is_package and not self.is_namespace:
            return self.full_path / "__init__.py"
        else:
            return self.full_path

    def iter_files_abs(self):
        """Iterate over the files contained in this module.
        Yields absolute paths. Excludes any __pycache__ and *.pyc files.
        """
        assert self.is_package or not self.is_namespace

        if self.is_generated:
            # Generated modules/packages don't exist in the source directory
            return

        def _include(s: str | Path):
            p = Path(s)
            return p.name != "__pycache__" and not p.name.endswith(".pyc")

        if self.is_package:
            # Ensure we sort all files and directories so the order is stable
            for dirpath, dirs, files in os.walk(str(self.full_path)):
                for file in sorted(files):
                    filepath = Path(dirpath) / file
                    if _include(filepath):
                        yield filepath
                dirs[:] = filter(_include, sorted(dirs))
        else:
            yield self.full_file


@dataclass
class Config:
    """Describes the metadata and configuration settings for the standard
    build backend."""

    standard_metadata: pyproject_metadata.StandardMetadata
    package_name: str = field(default="")
    module: dict[str, str | bool] = field(default_factory=dict)
    editable: dict[str, dict[str, Any]] = field(default_factory=dict)
    sdist: dict[str, dict[str, Any]] = field(default_factory=dict)
    cmake: dict[str, dict[str, Any]] | None = field(default=None)
    stubgen: dict[str, Any] | None = field(default=None)
    cross: dict[str, Any] | None = field(default=None)

    @property
    def referenced_files(self) -> list[Path]:
        metadata = self.standard_metadata
        res: list[Path] = []
        if metadata.readme is not None and metadata.readme.file is not None:
            res += [metadata.readme.file]
        if metadata.license is not None and metadata.license.file is not None:
            res += [metadata.license.file]
        return res

    def check(self):
        """Check for any incompatible options."""
        if self.module["namespace"] and any(
            e["mode"] == "wrapper" for e in self.editable.values()
        ):
            msg = "Namespace packages cannot use editable mode 'wrapper'"
            raise ConfigError(msg)
        if self.module.get("generated") and any(
            e["mode"] == "wrapper" for e in self.editable.values()
        ):
            msg = "Generated modules/packages cannot use editable mode 'wrapper'"
            raise ConfigError(msg)


@dataclass
class ComponentConfig:
    """Describes the metadata and configuration settings for the component
    build backend."""

    standard_metadata: pyproject_metadata.StandardMetadata
    package_name: str = field(default="")
    component: dict[str, Any] = field(default_factory=dict)


@dataclass
class PackageInfo:
    """Describes a Python package (names and version)."""

    version: str
    package_name: str
    module_name: str

    @property
    def norm_name(self):
        from .util import normalize_name_wheel

        return normalize_name_wheel(self.package_name)


@dataclass
class BuildPaths:
    """Paths used when building a (Wheel) package."""

    source_dir: Path  # Contains pyproject.toml
    build_dir: Path  # Usually .py-build-cmake_cache, determined by config
    wheel_dir: Path  # Where to place the Wheel when done (given by frontend)
    temp_dir: Path  # Temporary folder for use during build
    staging_dir: Path  # Where to install the Python modules and data files
    pkg_staging_dir: Path  # The folder that will be turned into a Wheel

    # Note that staging_dir and pkg_staging_dir can be different, e.g. when
    # doing an editable install, all Python files and modules are installed to
    # staging_dir, which is a subfolder of build_dir, and pkg_staging_dir only
    # contains a .pth file that points to staging_dir


@dataclass
class Command:
    args: Sequence[str]
    kwargs: dict[str, Any]


def format_and_rethrow_exception(e: BaseException):
    """Raises a FormattedErrorMessage from the given exception"""
    if isinstance(e, FormattedErrorMessage):
        raise e
    if isinstance(e, ConfigError):
        logger.error("Error in user configuration", exc_info=e)
        msg = "\n\n\t\u274C Error in user configuration:\n\n" f"\t\t{e}\n" "\n"
        raise FormattedErrorMessage(msg) from e
    if isinstance(e, CalledProcessError):
        logger.error("Subprocess failed", exc_info=e)
        if e.cmd[:2] == ["cmake", "--build"]:
            msg = "CMake build failed"
        elif e.cmd[:2] == ["cmake", "--install"]:
            msg = "CMake install failed"
        elif e.cmd[:1] == ["cmake"]:
            msg = "CMake failed"
        elif e.cmd[:1] == ["stubgen"]:
            msg = "Stub generation failed"
        else:
            msg = f"Subprocess {e.cmd[0]} failed"
        msg = (
            f"\n\n\t\u274C {msg}:\n\n"
            f"\t\t{e}\n"
            "\n\t(scroll up for subprocess output)"
        )
        raise FormattedErrorMessage(msg) from e
    elif isinstance(e, AssertionError):
        logger.error("Internal error:", exc_info=e)
        msg = (
            "\n"
            "\n"
            "\t\u274C Internal error:\n"
            "\t   Please notify the developers: https://github.com/tttapa/py-build-cmake/issues\n"
            "\n"
            f"\t\t{e}\n"
            "\n"
        )
        raise FormattedErrorMessage(msg) from e
    elif isinstance(e, KeyError):
        logger.error("Internal KeyError:", exc_info=e)
        msg = (
            "\n"
            "\n"
            "\t\u274C Internal KeyError:\n"
            "\t   Please notify the developers: https://github.com/tttapa/py-build-cmake/issues\n"
            "\n"
            f"\t\t{e}\n"
            "\n"
        )
        raise FormattedErrorMessage(msg) from e
    elif isinstance(e, Exception):
        logger.error("Uncaught exception:", exc_info=e)
        msg = (
            "\n"
            "\n"
            f"\t\u274C Uncaught exception: {type(e).__name__}\n"
            "\n"
            f"\t\t{e}\n"
            "\n"
        )
        raise FormattedErrorMessage(msg) from e



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path, PurePosixPath

from ...common import ConfigError
from .config_option import ConfigOption
from .config_path import ConfPath
from .default import DefaultValue
from .value_reference import ValueReference


@dataclass
class RelativeToCurrentConfig:
    project_path: Path | PurePosixPath
    description: str = "current configuration file"


@dataclass
class RelativeToProject:
    project_path: Path | PurePosixPath
    description: str = "project directory"


class PathConfigOption(ConfigOption):
    def __init__(
        self,
        name: str,
        description: str = "",
        example: str = "",
        default: DefaultValue | None = None,
        inherit_from: ConfPath | str | None = None,
        create_if_inheritance_target_exists: bool = False,
        must_exist: bool = True,
        expected_contents: list[str] | None = None,
        base_path: RelativeToProject | RelativeToCurrentConfig | None = None,
        allow_abs: bool = False,
        is_folder: bool = True,
    ):
        if expected_contents is None:
            expected_contents = []
        super().__init__(
            name,
            description,
            example,
            default,
            inherit_from,
            create_if_inheritance_target_exists,
        )
        self.must_exist = must_exist or bool(expected_contents)
        self.expected_contents = expected_contents
        self.base_path = base_path
        self.allow_abs = allow_abs
        self.is_folder = is_folder
        if self.base_path:
            assert self.base_path.project_path.is_absolute()

    def get_typename(self, md: bool = False):
        return "path" if self.is_folder else "filepath"

    def override(self, old_value, new_value):
        if new_value.values is None:
            return old_value.values
        return new_value.values

    def _verify_string(self, values: ValueReference):
        if self.sub_options:
            msg = f"Type of {values.value_path} should be {str}, "
            msg += f"not {dict}"
            raise ConfigError(msg)
        elif not isinstance(values.values, str):
            msg = f"Type of {values.value_path} should be {str}, "
            msg += f"not {type(values.values)}"
            raise ConfigError(msg)
        return values.values

    def check_path(self, values: ValueReference):
        assert isinstance(values.values, str)
        path: Path | PurePosixPath = Path(values.values)
        # Absolute or relative path?
        if path.is_absolute():
            # Absolute path
            if not self.allow_abs:
                msg = f'{values.value_path}: "{path!s}" must be a relative path'
                raise ConfigError(msg)
        # Relative path
        elif isinstance(self.base_path, RelativeToCurrentConfig):
            # value_path[0] is relative for files inside of the project,
            # otherwise it is absolute
            path = Path(PurePosixPath(values.value_path.pth[0]).parent / path)
            if not path.is_absolute():
                path = self.base_path.project_path / path
        elif isinstance(self.base_path, RelativeToProject):
            path = self.base_path.project_path / path
        else:
            msg = "Invalid relative path type"
            raise AssertionError(msg)
        assert path.is_absolute(), f"Failed to make path absolute: {path!s}"
        # Does the path exist?
        if self.must_exist:
            path = Path(path)
            if not path.exists():
                msg = f'{values.value_path}: "{path!s}" does not exist'
                raise ConfigError(msg)
            if self.is_folder != path.is_dir():
                type_ = "directory" if self.is_folder else "file"
                msg = f'{values.value_path}: "{path!s}" should be a {type_}'
                raise ConfigError(msg)
            # Are any of the required contents missing?
            missing = [
                sub for sub in self.expected_contents if not (path / Path(sub)).exists()
            ]
            if missing:
                missingstr = '", "'.join(missing)
                msg = f'{values.value_path}: "{path!s}" does not contain the following required files or folders: "{missingstr}"'
                raise ConfigError(msg)
        return path.resolve() if isinstance(path, Path) else path

    def verify(self, values: ValueReference):
        values.values = self._verify_string(values)
        if not values.values:
            return values.values
        return self.check_path(values)



from __future__ import annotations

import configparser
import contextlib
import logging
import os
import platform
import re
import sys
import sysconfig
from pathlib import Path
from typing import Any

from distlib.util import (  # type: ignore[import-untyped]
    get_platform as get_platform_dashes,
)

from ..common.util import (
    archflags_to_platform_tag,
    platform_to_platform_tag,
    python_sysconfig_platform_to_cmake_platform_win,
    python_sysconfig_platform_to_cmake_processor_win,
)
from .options.cmake_opt import CMakeOption
from .options.config_path import ConfPath
from .options.value_reference import ValueReference

logger = logging.getLogger(__name__)


def get_python_lib(library_dirs: str | list[str] | None) -> Path | None:
    """Return the path the the first python<major><minor>.lib or
    python<major>.lib file in any of the library_dirs.
    Returns None if no such file exists."""
    if library_dirs is None:
        return None
    if isinstance(library_dirs, str):
        library_dirs = [library_dirs]

    def possible_locations():
        v = sys.version_info
        py3xlib = lambda d: Path(d) / f"python{v.major}{v.minor}.lib"
        py3lib = lambda d: Path(d) / f"python{v.major}.lib"
        yield from map(py3xlib, library_dirs)
        yield from map(py3lib, library_dirs)

    try:
        return next(filter(Path.exists, possible_locations()))
    except StopIteration:
        return None


def cross_compile_win(
    config: ValueReference, plat_name, library_dirs, cmake_plat, cmake_proc
):
    """Update the configuration to include a cross-compilation configuration
    that builds for the given platform and processor. If library_dirs contains
    a compatible Python library, it is also included in the configuration, as
    well as the path to the Python installation's root directory, so CMake is
    able to locate Python correctly."""
    logger.info(
        "DIST_EXTRA_CONFIG.build_ext specified plat_name that is different from the current platform. "
        "Automatically enabling cross-compilation for %s",
        cmake_plat,
    )
    assert not config.is_value_set("cross")
    cross_cfg = {
        "os": "windows",
        "arch": platform_to_platform_tag(plat_name),
        "cmake": {
            "options": {
                "CMAKE_SYSTEM_NAME": CMakeOption.create("Windows", "STRING"),
                "CMAKE_SYSTEM_PROCESSOR": CMakeOption.create(cmake_proc, "STRING"),
                "CMAKE_GENERATOR_PLATFORM": CMakeOption.create(cmake_plat, "STRING"),
            }
        },
    }
    python_lib = get_python_lib(library_dirs)
    if python_lib is not None:
        cross_cfg["library"] = str(python_lib)
        python_root = python_lib.parent.parent
        if (python_root / "include").exists():
            cross_cfg["root"] = str(python_root)
    else:
        logger.warning(
            "Python library was not found in DIST_EXTRA_CONFIG.build_ext.library_dirs."
        )
    config.set_value("cross", cross_cfg)


def handle_cross_win(
    config: ValueReference, plat_name: str, library_dirs: str | list[str] | None
):
    """Try to configure cross-compilation for the given Windows platform.
    library_dirs should contain the directory with the Python library."""
    plat_proc = (
        python_sysconfig_platform_to_cmake_platform_win(plat_name),
        python_sysconfig_platform_to_cmake_processor_win(plat_name),
    )
    if all(plat_proc):
        cross_compile_win(config, plat_name, library_dirs, *plat_proc)
    else:
        logger.warning(
            "Cross-compilation setup skipped because the platform %s is unknown",
            plat_name,
        )


def handle_dist_extra_config_win(config: ValueReference, dist_extra_conf: str):
    """Read the given distutils configuration file and use it to configure
    cross-compilation if appropriate."""
    distcfg = configparser.ConfigParser()
    distcfg.read(dist_extra_conf)

    library_dirs = distcfg.get("build_ext", "library_dirs", fallback="")
    plat_name = distcfg.get("build_ext", "plat_name", fallback="")

    if plat_name and plat_name != sysconfig.get_platform():
        handle_cross_win(config, plat_name, library_dirs)


def config_quirks_win(config: ValueReference):
    """
    Explanation:
    The cibuildwheel tool sets the DIST_EXTRA_CONFIG environment variable when
    cross-compiling. It points to a configuration file that contains the path
    to the correct Python library (.lib), as well as the name of the platform
    to compile for.
    If the user did not specify a custom cross-compilation configuration,
    we will automatically add a minimal cross-compilation configuration that
    points CMake to the right Python library, and that selects the right
    CMake/Visual Studio platform.
    """
    dist_extra_conf = os.getenv("DIST_EXTRA_CONFIG")
    if dist_extra_conf is not None:
        if config.is_value_set("cross"):
            logger.warning(
                "Cross-compilation configuration was not empty, so I'm ignoring DIST_EXTRA_CONFIG"
            )
        elif not config.is_value_set("cmake"):
            logger.warning(
                "CMake configuration was empty, so I'm ignoring DIST_EXTRA_CONFIG"
            )
        else:
            handle_dist_extra_config_win(config, dist_extra_conf)


def cross_compile_mac(config: ValueReference, archs):
    """Update the configuration to include a cross-compilation configuration
    that builds for the given architectures."""
    logger.info(
        "ARCHFLAGS was specified. Automatically enabling cross-compilation for %s (native platform: %s)",
        ", ".join(archs),
        platform.machine(),
    )
    assert not config.is_value_set("cross")
    cross_cfg: dict[str, Any] = {
        "os": "mac",
        "cmake": {
            "options": {
                "CMAKE_SYSTEM_NAME": CMakeOption.create("Darwin", "STRING"),
                "CMAKE_OSX_ARCHITECTURES": CMakeOption.create(archs, "STRING"),
            }
        },
    }
    plat_tag = archflags_to_platform_tag(archs)
    if plat_tag:
        cross_arch = get_platform_dashes().split("-")
        cross_arch[-1] = plat_tag
        cross_cfg["arch"] = platform_to_platform_tag("_".join(cross_arch))
    if sys.implementation.name == "cpython":
        version = "".join(map(str, sys.version_info[:2]))
        abi = getattr(sys, "abiflags", "")
        env = cross_cfg["cmake"]["env"] = {}
        env["SETUPTOOLS_EXT_SUFFIX"] = f".cpython-{version}{abi}-darwin.so"
    config.set_value("cross", cross_cfg)


def config_quirks_mac(config: ValueReference):
    """Sets CMAKE_OSX_ARCHITECTURES if $ENV{ARCHFLAGS} is set
    on macOS. This ensures compatibility with cibuildwheel. If the interpreter
    architecture is not in the ARCHFLAGS, also enables cross-compilation."""
    archflags = os.getenv("ARCHFLAGS")
    if not archflags:
        return
    archs = list(re.findall(r"-arch +(\S+)", archflags))
    if not archs:
        logger.warning(
            "ARCHFLAGS was set, but its value was not valid, so I'm ignoring it"
        )
        return
    if config.is_value_set("cross"):
        logger.warning(
            "Cross-compilation configuration was not empty, so I'm ignoring ARCHFLAGS"
        )
        return
    if not config.is_value_set("cmake"):
        logger.warning("CMake configuration was empty, so I'm ignoring ARCHFLAGS")
        return
    if platform.machine() not in archs:
        cross_compile_mac(config, archs)
    else:
        logger.info(
            "ARCHFLAGS was set, adding CMAKE_OSX_ARCHITECTURES to cmake.options (%s, native platform: %s)",
            ", ".join(archs),
            platform.machine(),
        )
        config.set_value_default(ConfPath.from_string("cmake/options"), {})
        config.set_value_default(
            ConfPath.from_string("cmake/options/CMAKE_OSX_ARCHITECTURES"),
            CMakeOption.create(archs, "STRING"),
        )


def config_quirks_pypy(config: ValueReference):
    if sys.version_info < (3, 8):
        with contextlib.suppress(KeyError):
            del config.values["stubgen"]
            logger.info("Mypy is not supported on PyPy <3.8, disabling stubgen")


def config_quirks(config: ValueReference):
    dispatch = {
        "Windows": config_quirks_win,
        "Darwin": config_quirks_mac,
    }.get(platform.system())
    if dispatch is not None:
        dispatch(config)
    dispatch = {
        "pypy": config_quirks_pypy,
    }.get(sys.implementation.name)
    if dispatch is not None:
        dispatch(config)



from __future__ import annotations

from copy import deepcopy

from ...common import ConfigError
from .config_option import ConfigOption
from .value_reference import ValueReference


class DictOfStrConfigOption(ConfigOption):
    def get_typename(self, md: bool = False) -> str:
        return "dict"

    def override(self, old_value: ValueReference, new_value: ValueReference):
        if old_value.values is None:
            old_value.values = {}
        if new_value.values is None:
            return old_value.values
        r = deepcopy(old_value.values)
        r.update(deepcopy(new_value.values))
        return r

    def verify(self, values: ValueReference):
        if values.values is None:
            return None
        valdict = values.values
        if not isinstance(valdict, dict):
            msg = f"Type of {values.value_path} should be {dict}, "
            msg += f"not {type(valdict)}"
            raise ConfigError(msg)
        elif not all(isinstance(el, str) for el in valdict):
            msg = f"Type of keys in {values.value_path} should be {str}"
            raise ConfigError(msg)
        elif not all(isinstance(el, str) for el in valdict.values()):
            msg = f"Type of values in {values.value_path} should be {str}"
            raise ConfigError(msg)
        return valdict



from __future__ import annotations


class ConfPath:
    def __init__(self, pth: tuple[str, ...] = ()):
        self.pth = pth

    @classmethod
    def from_string(cls, str_pth: str):
        return cls(tuple(filter(None, str_pth.split("/"))))

    def __str__(self):
        return "/".join(".." if x == "^" else x for x in self.pth)

    def join(self, suffix: str | ConfPath):
        if isinstance(suffix, str):
            suffix = ConfPath((suffix,))
        p1 = self.pth
        p2 = suffix.pth
        while p1 and p2 and p2[0] == "^":
            p1 = p1[:-1]
            p2 = p2[1:]
        return ConfPath(p1 + p2)

    def split(self) -> tuple[ConfPath, str]:
        return ConfPath(self.pth[:-1]), self.pth[-1]

    def split_front(self) -> tuple[str, ConfPath]:
        return self.pth[0], ConfPath(self.pth[1:])

    def __bool__(self):
        return bool(self.pth)



from __future__ import annotations

import html
import itertools
import re
import shutil
import textwrap

from .config.options.config_option import ConfigOption
from .config.options.default import RequiredValue
from .config.options.path import PathConfigOption


def _print_wrapped(text: str, indent, width: int | None = None):
    """Print the given string with the given indentation, wrapping it to the
    desired width, preserving line endings. If `width` is None, use the width
    of the terminal, or 80 columns as a fallback."""
    if width is None:
        width = shutil.get_terminal_size((80, 24)).columns
    wrapper = textwrap.TextWrapper(
        width=width, initial_indent=indent, subsequent_indent=indent
    )
    wrapped = [wrapper.wrap(i) for i in text.split("\n")]
    for line in itertools.chain.from_iterable(wrapped):
        print(line)


def get_default_str(opt: ConfigOption):
    """Get a string representation of the default value for the given option."""
    return opt.default.get_name()


def help_print_md(pbc_opts: ConfigOption):
    """
    Prints the top-level options in `pbc_opts` as MarkDown tables.
    """
    for k, v in pbc_opts.sub_options.items():
        print("##", k)
        print(_get_full_description(v), "\n")
        print("| Option | Description | Type | Default |")
        print("|--------|-------------|------|---------|")
        for kk, vv in v.sub_options.items() or {}:
            typename = vv.get_typename(md=True) or ""
            print(
                "|",
                f"`{kk}`",
                "|",
                _get_full_description(vv),
                "|",
                typename,
                "|",
                f"`{get_default_str(vv)}`",
                "|",
            )
        print()


def _get_full_description(vv: ConfigOption):
    descr = _md_escape(vv.description)
    if isinstance(vv, PathConfigOption):
        descr += "<br/>" + _describe_path_option(vv).capitalize() + "."
    if vv.inherits:
        descr += "<br/>Inherits from: `/" + str(vv.inherits) + "`"
    if vv.example:
        descr += "<br/>For example: `" + vv.example + "`"
    return descr


def _md_escape(descr: str):
    descr = html.escape(descr)
    descr = descr.replace("*", "\\*")
    descr = descr.replace("_", "\\_")

    def unescape(m: re.Match):
        m0 = m.group(0)
        return html.unescape(m0.replace("\\_", "_").replace("\\*", "*"))

    descr = re.sub(r"`.*`", unescape, descr)
    return descr.replace("\n", "<br/>")


def recursive_help_print(opt: ConfigOption, level=0):
    """Recursively prints the help messages for the options in `opt`."""
    for k, v in opt.sub_options.items():
        if k == "project":
            continue
        indent = 4 * level * " "
        header = "\n" + k
        if v.sub_options:
            header += ":"
            print(textwrap.indent(header, indent))
            if v.description:
                _print_wrapped(v.description, indent + "  ")
            recursive_help_print(v, level + 1)
        else:
            headerfields = []
            typename = v.get_typename()
            if typename is not None:
                headerfields += [typename]
            if isinstance(v, PathConfigOption):
                headerfields += [_describe_path_option(v)]
            is_required = isinstance(v.default, RequiredValue)
            if is_required:
                headerfields += ["required"]
            if v.inherits:
                headerfields += ["inherits from /" + str(v.inherits)]
            if headerfields:
                header += " (" + ", ".join(headerfields) + ")"
            print(textwrap.indent(header + ":", indent))
            _print_wrapped(v.description, indent + "  ")
            if v.example:
                _print_wrapped("For example: " + v.example, indent + "  ")
            default = v.default.get_name()
            if not is_required:
                print(textwrap.indent("Default: " + default, indent + "  "))


def _describe_path_option(v: PathConfigOption):
    t = "absolute or relative" if v.allow_abs else "relative"
    if v.base_path is not None:
        t += " to " + v.base_path.description
    return t


def help_print(pbc_opts: ConfigOption):
    recursive_help_print(pbc_opts)
    print()



from __future__ import annotations

from pathlib import Path

import click

from . import __version__


def cmake_command(directory, build_path, verbose, dry, native, cross, local):
    def get_cmaker():
        from .build import _BuildBackend as backend
        from .commands.cmd_runner import CommandRunner

        src_dir = Path(directory or ".").resolve()
        config_settings = {
            "--cross": list(cross),
            "--local": list(local),
        }
        # Read configuration and package metadata
        cfg, module = backend.read_all_metadata(src_dir, config_settings, verbose)
        pkg_info = backend.get_pkg_info(cfg, module)
        cmake_cfg = backend.get_cmake_config(cfg)

        # Set up all paths
        build_cfg_name = backend.get_build_config_name(cfg.cross)
        build_dir = Path(cmake_cfg["build_path"]) / build_cfg_name

        # CMake builder
        return backend.get_cmaker(
            src_dir,
            build_dir,
            None,
            cmake_cfg,
            cfg.cross,
            pkg_info,
            runner=CommandRunner(verbose=verbose, dry=dry),
        )

    return get_cmaker


@click.group()
@click.option(
    "-C",
    "--directory",
    type=click.Path(exists=True, file_okay=False, dir_okay=True),
    required=False,
    help="The directory containing pyproject.toml.",
)
@click.option(
    "-B",
    "--build-path",
    type=click.Path(exists=False, file_okay=False, dir_okay=True),
    required=False,
    default=None,
    help="Override py-build-cmake's default build folder.",
)
@click.option(
    "-V",
    "--verbose",
    is_flag=True,
    help="Print verbose information about the commands being executed.",
)
@click.option(
    "-n",
    "--dry",
    is_flag=True,
    help="Print the commands without actually invoking CMake.",
)
@click.option(
    "--native",
    is_flag=True,
    help="When the configuration requests cross-compiling, "
    "configure for a native build instead. Has no effect otherwise.",
)
@click.option(
    "--local",
    type=click.Path(exists=False, file_okay=True, dir_okay=False),
    required=False,
    multiple=True,
    help="Specifies a toml file that overrides the "
    "tool.py-build-cmake section of pyproject.toml, "
    "similar to py-build-cmake.local.toml.",
)
@click.option(
    "--cross",
    type=click.Path(exists=False, file_okay=True, dir_okay=False),
    required=False,
    multiple=True,
    help="Specifies a toml file that overrides the "
    "tool.py-build-cmake.cross section of pyproject.toml, "
    "similar to py-build-cmake.cross.toml.",
)
@click.version_option(__version__, "-v", "--version", prog_name="py-build-cmake")
@click.pass_context
def cli(ctx: click.Context, **kwargs):
    ctx.obj = cmake_command(**kwargs)


@cli.command(help="Configure the CMake project.")
@click.pass_obj
@click.option(
    "--preset",
    nargs=1,
    type=str,
    required=False,
    metavar="PRESET",
    help="CMake configure preset to use.",
)
@click.option(
    "--use-build-presets",
    is_flag=True,
    help="Indicate that a build preset will be used during the build stage. "
    "This causes py-build-cmake to let CMake pick the build directory during "
    "configuration, rather than explicitly overriding it.",
)
@click.argument("args", nargs=-1, required=False)
def configure(obj, preset, use_build_presets, args):
    cmaker = obj()
    if cmaker is None:
        return
    cmaker.conf_settings.args += args or []
    if preset is not None:
        cmaker.conf_settings.preset = preset
    if use_build_presets:
        cmaker.build_settings.presets = [""]
    cmaker.configure()


@cli.command(help="Build the CMake project.")
@click.pass_obj
@click.option(
    "--preset", nargs=1, multiple=True, type=str, required=False, metavar="PRESET"
)
@click.option(
    "--config", nargs=1, multiple=True, type=str, required=False, metavar="CONFIG"
)
@click.argument("args", nargs=-1, required=False)
def build(obj, preset, config, args):
    cmaker = obj()
    if cmaker is None:
        return
    cmaker.build_settings.args += args or []
    if preset or config:
        cmaker.build_settings.presets = []
        cmaker.build_settings.configs = []
    if preset:
        cmaker.build_settings.presets = preset
    if config:
        cmaker.build_settings.configs = config
    cmaker.build()


@cli.command(help="Install the CMake project.")
@click.pass_obj
@click.option(
    "--config", nargs=1, multiple=True, type=str, required=False, metavar="CONFIG"
)
@click.option(
    "--component", nargs=1, multiple=True, type=str, required=False, metavar="COMP"
)
@click.argument("args", nargs=-1, required=False)
def install(obj, config, component, args):
    cmaker = obj()
    if cmaker is None:
        return
    cmaker.install_settings.args += args or []
    if config:
        cmaker.install_settings.configs = config
    if component:
        cmaker.install_settings.components = component
    cmaker.install()


@cli.group(help="Config file operations.")
def config():
    pass


@config.command(help="Print documentation for the config file format.")
@click.option("--md", is_flag=True, help="Use the MarkDown format.")
@click.option(
    "--component", is_flag=True, help="Documentation for the build_component backend."
)
def format(md, component):
    from .config.options.config_path import ConfPath
    from .config.options.config_reference import ConfigReference
    from .config.options.pyproject_options import get_component_options, get_options

    if md:
        from .help import help_print_md as help_print
    else:
        from .help import help_print
    help_pth = ConfPath.from_string("pyproject.toml/tool/py-build-cmake")
    pr_md = print if md else lambda *args, **kwargs: None
    pr_tx = lambda *args, **kwargs: None if md else print
    if component:
        pr_tx(
            "List of py-build-cmake pyproject.toml options for the "
            "build_component backend:"
        )
        pr_md(
            "# py-build-cmake component build backend\n"
            "The `py_build_cmake.build_component` build backend allows "
            "building packages containing additional binaries that are not "
            "included with the main distribution.\n"
        )
        pr_md(
            "A possible use case is distributing debug symbols: these files "
            "can be large, and most users don't need them, so distributing "
            "them in a separate package makes sense.\n"
        )
        pr_md(
            "See [examples/minimal-debug-component](https://github.com/tttapa/"
            "py-build-cmake/tree/main/examples/minimal-debug-component) "
            "for more information.\n"
        )
        opts = get_component_options(Path("/"))
        root_ref = ConfigReference(ConfPath.from_string("/"), opts)
        help_print(root_ref.sub_ref(help_pth).config)

    else:
        pr_tx("List of py-build-cmake pyproject.toml options:")
        pr_md("# py-build-cmake configuration options\n")
        pr_md(
            "These options go in the `[tool.py-build-cmake]` section of "
            "the `pyproject.toml` configuration file.\n"
        )
        opts = get_options(Path("/"))
        root_ref = ConfigReference(ConfPath.from_string("/"), opts)
        help_print(root_ref.sub_ref(help_pth).config)
        pr_md("# Local overrides\n")
        pr_md(
            "Additionally, two extra configuration files can be placed in "
            "the same directory as `pyproject.toml` to override some "
            "options for your specific use case:\n\n"
            "- `py-build-cmake.local.toml`: the options in this file "
            "override the values in the `tool.py-build-cmake` section of "
            "`pyproject.toml`.<br/>This is useful if you need specific "
            "arguments or CMake options to compile the package on your "
            "system.\n"
            "- `py-build-cmake.cross.toml`: the options in this file "
            "override the values in the `tool.py-build-cmake.cross` section "
            "of `pyproject.toml`.<br/>Useful for cross-compiling the "
            "package without having to edit the main configuration file.\n"
        )
        pr_md("# Command line overrides\n")
        pr_md(
            "Instead of using the `py-build-cmake.local.toml` and "
            "`py-build-cmake.cross.toml` files, you can also include "
            "additional config files using command line options:\n\n"
            "- `--local`: specifies a toml file that overrides the "
            "`tool.py-build-cmake` section of `pyproject.toml`, "
            "similar to `py-build-cmake.local.toml`\n"
            "- `--cross`: specifies a toml file that overrides the "
            "`tool.py-build-cmake.cross` section of `pyproject.toml`, "
            "similar to `py-build-cmake.cross.toml`\n\n"
            "These command line overrides are applied after the "
            "`py-build-cmake.local.toml` and `py-build-cmake.cross.toml` "
            "files in the project folder (if any).\n\n"
            "When using PyPA build, these flags can be specified using "
            "the `-C` or `--config-setting` flag: \n"
            "```sh\n"
            "python -m build . -C--cross=/path/to/my-cross-config.toml\n"
            "```\n"
            "The same flag may appear multiple times, for example: \n"
            "```sh\n"
            "python -m build . -C--local=conf-A.toml -C--local=conf-B.toml\n"
            "```\n"
            "For PyPA pip, you can use the `--config-settings` flag "
            "instead.\n"
        )
        pr_md("# Combining lists\n")
        pr_md(
            "Many options are specified as lists of strings. When one of "
            "these options inherits from or is overridden by another option, "
            "these lists can be merged in different ways.\n\n"
            "In the table above, when the data type is `list`, the original "
            "list of options is simply replaced by the list it is overridden "
            "by.  \n"
            "When the data type is `list+`, the value of the overriding "
            "option is appended to the original value. This is used primarily "
            "for combining lists of command line options.\n\n"
            "If you want full control of what happens when overriding a list "
            "option, you can use a dictionary with one or more of the "
            "following keys:\n\n"
            '  - `"="`: replace the original list by this list (default '
            "behavior for options of type `list`)\n"
            '  - `"+"`: append the values of this list to the end of the '
            "original list (default behavior for options of type `list+`)\n"
            '  - `"-"`: remove the values in this list from the original '
            "list (if present)\n"
            '  - `"prepend"`: prepend the values of this list to the '
            "beginning of the original list\n\n"
            "Some examples:\n"
            """
```toml
[cmake]
build_args = ["--verbose", "--clean-first"]
[linux.cmake]
build_args = ["--target", "foo"]
[windows.cmake]
build_args = {"-" = ["--verbose"], "+" = ["--target", "bar"]}
[macos.cmake]
build_args = {"=" = ["--target", "macos"]}
```
"""
            "Because `linux.cmake` inherits from `cmake`, and because "
            "`build_args` has type `list+`, this will result in a value of "
            '`["--verbose", "--clean-first", "--target", "foo"]` for the '
            "`linux.cmake.build_args` option. The value of "
            "`windows.cmake.build_args` will be "
            '`["--clean_first", "--target", "bar"]`, and the value of '
            '`macos.cmake.build_args` will be `["--target", "macos"]`.\n'
            """
```toml
[cmake]
config = ["Debug", "Release"]
[linux.cmake]
config = ["RelWithDebInfo"]
[windows.cmake]
config = {"prepend" = ["RelWithDebInfo"], "-" = ["Debug"], "+" = ["Debug"]}
```
"""
            "The `build_args` option has type `list`, so the value of "
            '`linux.cmake.config` is simply `["RelWithDebInfo"]`. The value '
            "of `windows.cmake.config` is "
            '`["RelWithDebInfo", "Release", "Debug"]`.\n\n'
            "The same rules also apply to CMake options:\n"
            """
```toml
[cmake.options]
CMAKE_PREFIX_PATH = "/some/path"
[linux.cmake.options]
CMAKE_PREFIX_PATH = {"prepend" = "/some/linux-specific/path"}
```
"""
            "This passes the option "
            "`-D CMAKE_PREFIX_PATH=/some/linux-specific/path;/some/path` to "
            "CMake."
        )


if __name__ == "__main__":
    cli()



from __future__ import annotations

import shutil
import textwrap
from pathlib import Path

from ...common import ConfigError, Module


def write_editable_wrapper(staging_dir: Path, module: Module):
    """Write a fake __init__.py file that points to the development folder."""
    if not module.is_package:
        msg = "Editable wrapper mode is not supported for stand-alone Python modules."
        raise ConfigError(msg)
    name = module.name
    tmp_pkg = staging_dir / name
    pkgpath = module.full_path
    initpath = pkgpath / "__init__.py"
    tmp_pkg.mkdir(parents=True, exist_ok=True)
    special_dunders = [
        "__builtins__",
        "__cached__",
        "__file__",
        "__loader__",
        "__name__",
        "__package__",
        "__path__",
        "__spec__",
    ]
    content = f"""\
        # First extend the search path with the development folder
        __spec__.submodule_search_locations.insert(0, {str(pkgpath)!a})
        # Now manually import the development __init__.py
        from importlib import util as _util
        _spec = _util.spec_from_file_location("{name}",
                                                {str(initpath)!a})
        _mod = _util.module_from_spec(_spec)
        _spec.loader.exec_module(_mod)
        # After importing, add its symbols to our global scope
        _vars = _mod.__dict__.copy()
        for _k in ['{"','".join(special_dunders)}']: _vars.pop(_k)
        globals().update(_vars)
        # Clean up
        del _k, _spec, _mod, _vars, _util
        """
    (tmp_pkg / "__init__.py").write_text(textwrap.dedent(content), encoding="utf-8")
    # Add the py.typed file if it exists, so mypy picks up the stubs for
    # the C++ extensions
    py_typed = module.full_path / "py.typed"
    if py_typed.exists():
        shutil.copy2(py_typed, tmp_pkg)
    # Write a path file so IDEs find the correct files as well
    (staging_dir / f"{name}.pth").write_text(f"{module.full_path.parent!s}\n")



from __future__ import annotations

from ...common import BuildPaths, Config, Module
from ...common.util import get_os_name
from .hook import write_editable_hook
from .symlink import write_editable_links
from .wrapper import write_editable_wrapper


def do_editable_install(cfg: Config, paths: BuildPaths, module: Module):
    edit_cfg = cfg.editable["cross" if cfg.cross else get_os_name()]
    mode = edit_cfg["mode"]
    if mode == "wrapper":
        write_editable_wrapper(paths.staging_dir, module)
    elif mode == "hook":
        write_editable_hook(paths.staging_dir, module)
    elif mode == "symlink":
        paths = write_editable_links(paths, module)
    else:
        msg = "Invalid editable mode"
        raise AssertionError(msg)
    return paths



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .config_option import ConfigOption
from .config_path import ConfPath


class ConfigReference:
    def __init__(self, config_path: ConfPath, config: ConfigOption) -> None:
        self.config_path = config_path
        self.config = config

    @property
    def sub_options(self):
        return self.config.sub_options

    def sub_ref(self, name: str | ConfPath) -> ConfigReference:
        if isinstance(name, ConfPath) and len(name.pth) == 1:
            name = name.pth[0]
        if isinstance(name, str):
            return ConfigReference(
                config_path=self.config_path.join(name),
                config=self.config.sub_options[name],
            )
        name, rem = name.split_front()
        return self.sub_ref(name).sub_ref(rem)

    def resolve_inheritance_single(self, root: ConfigReference) -> ConfigReference:
        if self.config.inherits is None:
            return self
        # Find the option we inherit from and make sure it exists
        parent_path = self.config.inherits
        # TODO: should we support relative paths here?
        # self.config_path.join(self.config.inherits)
        try:
            parent = root.sub_ref(parent_path)
        except KeyError as e:
            msg = f"Inheritance target {parent_path} is not a valid option"
            msg += f" (in {self.config_path})"
            raise ValueError(msg) from e
        return parent

    def resolve_inheritance(self, root: ConfigReference) -> ConfigReference:
        if self.config.inherits is None:
            return self
        return self.resolve_inheritance_single(root).resolve_inheritance(root)



from __future__ import annotations

import logging
import os
from pathlib import Path


class GitHubActionsFormatter(logging.Formatter):
    """Formats warnings etc. for GitHub Actions: https://docs.github.com/en/actions/using-workflows/workflow-commands-for-github-actions#setting-a-notice-message"""

    def __init__(self):
        super().__init__(fmt="%(name)s:%(message)s")

    def format(self, record: logging.LogRecord):
        s = super().format(record)
        loc = ""
        if os.environ.get("GITHUB_REPOSITORY", "").endswith("/py-build-cmake"):
            p = Path(record.pathname)
            try:
                i = len(p.parts) - p.parts[-1::-1].index("py_build_cmake") - 1
                file = Path("src", *p.parts[i:])
                loc = f" file={file},line={record.lineno}"
            except ValueError:
                pass
        prefix = {
            logging.INFO: f"::notice{loc}::",
            logging.WARNING: f"::warning{loc}::",
            logging.ERROR: f"::error{loc}::",
        }.get(record.levelno, record.levelname + ":")
        return prefix + s



from __future__ import annotations

import os.path as osp
import re
from copy import copy
from pathlib import PurePath

from ...common import ConfigError
from .config_path import ConfPath
from .default import DefaultValue
from .list import ListOfStrConfigOption
from .value_reference import ValueReference


class DirPatternsConfigOption(ListOfStrConfigOption):
    def __init__(
        self,
        name: str,
        description: str = "",
        example: str = "",
        default: DefaultValue | None = None,
        inherit_from: ConfPath | str | None = None,
        create_if_inheritance_target_exists: bool = False,
        convert_str_to_singleton=False,
    ) -> None:
        super().__init__(
            name,
            description,
            example,
            default,
            inherit_from,
            create_if_inheritance_target_exists,
            convert_str_to_singleton,
        )

    def verify(self, values: ValueReference):
        # Based on https://github.com/pypa/flit/blob/f7496a50debdfa393e39f8e51d328deabcd7ae7e/flit_core/flit_core/config.py#L215
        super().verify(values)
        # Windows filenames can't contain these (nor * or ?, but they are part of
        # glob patterns) - https://stackoverflow.com/a/31976060/434217
        bad_chars = re.compile(r'[\000-\037<>:"\\]')
        pattern_list = copy(values.values)
        for i, pattern in enumerate(pattern_list):
            if bad_chars.search(pattern):
                msg = f"Pattern '{pattern}' in {values.value_path} contains bad characters (<>:\"\\ or control characters)"
                raise ConfigError(msg)
            # Normalize the path
            normp = PurePath(osp.normpath(pattern))
            # Make sure that the path is relative and inside of the project
            if normp.is_absolute():
                msg = f"Pattern '{pattern}' in {values.value_path} should be relative"
                raise ConfigError(msg)
            if normp.parts[0] == "..":
                msg = f"Pattern '{pattern}' in {values.value_path} cannot refer to the parent directory (..)"
                raise ConfigError(msg)
            pattern_list[i] = str(normp)
        return pattern_list



from __future__ import annotations

from copy import copy

from .config_reference import ConfigReference
from .value_reference import ValueReference


class ConfigVerifier:
    def __init__(
        self, root: ConfigReference, ref: ConfigReference, values: ValueReference
    ) -> None:
        self.root = root
        self.ref = ref
        self.values = values

    def verify(self):
        verified_values = ValueReference(
            self.values.value_path,
            self.ref.config.verify(copy(self.values)),
        )
        for name in self.ref.sub_options:
            if name in self.values.values:
                verified_values.values[name] = ConfigVerifier(
                    root=self.root,
                    ref=self.ref.sub_ref(name).resolve_inheritance(self.root),
                    values=verified_values.sub_ref(name),
                ).verify()
        return verified_values.values



from __future__ import annotations

from ...common import ConfigError
from .config_option import ConfigOption
from .value_reference import ValueReference


class IntConfigOption(ConfigOption):
    def get_typename(self, md: bool = False):
        return "int"

    def override(self, old_value: ValueReference, new_value: ValueReference):
        if new_value.values is None:
            return old_value.values
        return new_value.values

    def verify(self, values: ValueReference):
        if self.sub_options:
            msg = f"Type of {values.value_path} should be {int}, "
            msg += f"not {dict}"
            raise ConfigError(msg)
        elif not isinstance(values.values, int):
            msg = f"Type of {values.value_path} should be {int}, "
            msg += f"not {type(values.values)}"
            raise ConfigError(msg)
        return values.values



from __future__ import annotations

from copy import copy
from difflib import get_close_matches
from typing import Any, Iterable

from ...common import ConfigError
from .config_path import ConfPath
from .default import DefaultValue, NoDefaultValue
from .value_reference import ValueReference


class ConfigOption:
    def __init__(
        self,
        name: str,
        description: str = "",
        example: str = "",
        default: DefaultValue | None = None,
        inherit_from: ConfPath | str | None = None,
        create_if_inheritance_target_exists: bool = False,
    ) -> None:
        if default is None:
            default = NoDefaultValue()
        if isinstance(inherit_from, str):
            inherit_from = ConfPath.from_string(inherit_from)
        self.name = name
        self.description = description
        self.example = example
        self.sub_options: dict[str, ConfigOption] = {}
        self.inherits = inherit_from
        self.default: DefaultValue = default
        self.create_if_inheritance_target_exists = create_if_inheritance_target_exists

    def insert(self, opt: ConfigOption):
        assert opt.name not in self.sub_options
        self.sub_options[opt.name] = opt
        return self.sub_options[opt.name]

    def insert_multiple(self, opts: Iterable[ConfigOption]):
        for opt in opts:
            self.insert(opt)
        return self

    def get_typename(self, md: bool = False) -> str | None:
        return None

    def verify(self, values: ValueReference) -> Any:
        if not isinstance(values.values, dict):
            msg = f"Type of {values.value_path} should be 'dict', "
            msg += f"not {type(values.values)}"
            raise ConfigError(msg)
        unknown_keys = set(values.values) - set(self.sub_options)
        msg = ""
        for k in sorted(unknown_keys):
            suggested = get_close_matches(k, self.sub_options, 3)
            msg = f"Unknown option '{k}' in {values.value_path}. "
            msg += f"Did you mean: {', '.join(suggested)}\n"
        if msg:
            raise ConfigError(msg[:-1])
        return values.values

    def finalize(self, values: ValueReference) -> Any:
        return values.values

    def override(self, old_value: ValueReference, new_value: ValueReference) -> Any:
        return copy(old_value.values)


class UncheckedConfigOption(ConfigOption):
    def verify(self, values: ValueReference):
        return values.values



from __future__ import annotations

import platform
import re
import sys
from typing import Sequence, cast

if sys.version_info < (3, 8):
    OSIdentifier = str
else:
    from typing import Literal

    OSIdentifier = Literal["linux", "windows", "mac"]


def get_os_name() -> OSIdentifier:
    """Get the name of the current platform."""
    osname = {
        "Linux": "linux",
        "Windows": "windows",
        "Darwin": "mac",
    }.get(platform.system())
    if not osname:
        msg = "Unsupported platform"
        raise ValueError(msg)
    return cast(OSIdentifier, osname)


def normalize_name_wheel_pep_427(name: str) -> str:
    """https://www.python.org/dev/peps/pep-0427/#escaping-and-unicode"""
    return re.sub(r"[^\w\d.]+", "_", name, flags=re.UNICODE)


def normalize_name_wheel(name: str) -> str:
    """https://packaging.python.org/en/latest/specifications/binary-distribution-format/#escaping-and-unicode"""
    return re.sub(r"[-_.]+", "_", name).lower()


def python_sysconfig_platform_to_cmake_platform_win(
    plat_name: str | None,
) -> str | None:
    """Convert a sysconfig platform string to the corresponding value of
    https://cmake.org/cmake/help/latest/variable/CMAKE_GENERATOR_PLATFORM.html"""
    return {
        None: None,
        "win32": "Win32",
        "win-amd64": "x64",
        "win-arm32": "ARM",
        "win-arm64": "ARM64",
    }.get(plat_name)


def python_sysconfig_platform_to_cmake_processor_win(
    plat_name: str | None,
) -> str | None:
    """Convert a sysconfig platform string to the corresponding value of
    https://cmake.org/cmake/help/latest/variable/CMAKE_HOST_SYSTEM_PROCESSOR.html"""
    # The value of %PROCESSOR_ARCHITECTURE% on Windows
    return {
        None: None,
        "win32": "x86",
        "win-amd64": "AMD64",
        "win-arm32": "ARM",
        "win-arm64": "ARM64",
    }.get(plat_name)


def platform_to_platform_tag(plat: str) -> str:
    """https://packaging.python.org/en/latest/specifications/platform-compatibility-tags/#platform-tag"""
    return plat.replace(".", "_").replace("-", "_")


def archflags_to_platform_tag(archflags: Sequence[str]) -> str | None:
    """Convert tuple of CMAKE_OSX_ARCHITECTURES values to the corresponding
    platform tag https://packaging.python.org/en/latest/specifications/platform-compatibility-tags/#platform-tag
    """
    return {
        ("arm64", "x86_64"): "universal2",
        ("x86_64",): "x86_64",
        ("arm64",): "arm64",
    }.get(tuple(sorted(archflags)))



from __future__ import annotations

import shutil
from copy import copy

from ...common import BuildPaths, Module
from ..util import copy_pkg_source_to


def write_editable_links(paths: BuildPaths, module: Module) -> BuildPaths:
    paths = copy(paths)
    cache_dir = paths.build_dir
    cache_dir.mkdir(exist_ok=True, parents=True)
    paths.staging_dir = cache_dir / "editable"
    shutil.rmtree(paths.staging_dir, ignore_errors=True)
    paths.staging_dir.mkdir()
    copy_pkg_source_to(paths.staging_dir, module, symlink=True)
    pth_file = paths.pkg_staging_dir / f"{module.name}.pth"
    pth_file.parent.mkdir(exist_ok=True)
    pth_file.write_text(f"{paths.staging_dir!s}\n")
    return paths



"""
The following functions are based on flit_core, under the BSD 3-Clause license:

Copyright (c) 2015, Thomas Kluyver and contributors
All rights reserved.

BSD 3-clause license:

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its contributors
may be used to endorse or promote products derived from this software without
specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from __future__ import annotations

import ast
import logging
import sys
from contextlib import contextmanager
from pathlib import Path

import distlib.version  # type: ignore[import-untyped]

from ..common import (
    ConfigError,
    InvalidVersion,
    Module,
    NoDocstringError,
    NoVersionError,
    ProblemInModule,
)

logger = logging.getLogger(__name__)


@contextmanager
def _module_load_ctx():
    """Preserve some global state that modules might change at import time.

    - Handlers on the root logger.
    """
    logging_handlers = logging.root.handlers[:]
    try:
        yield
    finally:
        logging.root.handlers = logging_handlers


def get_docstring_and_version_via_ast(mod_filename: Path):
    """
    Return a tuple like (docstring, version) for the given module,
    extracted by parsing its AST.
    """
    # read as bytes to enable custom encodings
    with mod_filename.open("rb") as f:
        node = ast.parse(f.read())
    version = None
    for child in node.body:
        # Only use the version from the given module if it's a simple
        # string assignment to __version__
        if (
            isinstance(child, ast.Assign)
            and any(
                isinstance(target, ast.Name) and target.id == "__version__"
                for target in child.targets
            )
            and isinstance(child.value, ast.Str)
        ):
            version = child.value.s
            break
    return ast.get_docstring(node), version


# To ensure we're actually loading the specified file, give it a unique name to
# avoid any cached import. In normal use we'll only load one module per process,
# so it should only matter for the tests, but we'll do it anyway.
_import_i = 0


def get_docstring_and_version_via_import(mod_filename: Path):
    """
    Return a tuple like (docstring, version) for the given module,
    extracted by importing the module and pulling __doc__ & __version__
    from it.
    """
    global _import_i  # noqa: PLW0603
    _import_i += 1

    logger.debug("Loading module %s", mod_filename)
    from importlib.util import module_from_spec, spec_from_file_location

    mod_name = "py_build_cmake.dummy.import%d" % _import_i
    spec = spec_from_file_location(mod_name, mod_filename)
    if spec is None:
        msg = f"Unable to import '{mod_filename}' (missing spec)"
        raise ProblemInModule(msg)
    if spec.loader is None:
        msg = f"Unable to import '{mod_filename}' (missing loader)"
        raise ProblemInModule(msg)
    with _module_load_ctx():
        m = module_from_spec(spec)
        # Add the module to sys.modules to allow relative imports to work.
        # importlib has more code around this to handle the case where two
        # threads are trying to load the same module at the same time, but Flit
        # should always be running a single thread, so we won't duplicate that.
        sys.modules[mod_name] = m
        try:
            spec.loader.exec_module(m)
        finally:
            sys.modules.pop(mod_name, None)

    docstring = m.__dict__.get("__doc__", None)
    version = m.__dict__.get("__version__", None)
    return docstring, version


def get_info_from_module(mod_filename: Path, for_fields=("version", "description")):
    """Load the module/package, get its docstring and __version__"""
    if not for_fields:
        return {}

    # What core metadata calls Summary, PEP 621 calls description
    want_summary = "description" in for_fields
    want_version = "version" in for_fields

    logger.debug("Loading module %s", mod_filename)

    # Attempt to extract our docstring & version by parsing our target's
    # AST, falling back to an import if that fails. This allows us to
    # build without necessarily requiring that our built package's
    # requirements are installed.
    docstring, version = get_docstring_and_version_via_ast(mod_filename)
    if (want_summary and not docstring) or (want_version and not version):
        docstring, version = get_docstring_and_version_via_import(mod_filename)

    res = {}

    if want_summary:
        if (not docstring) or not docstring.strip():
            msg = f"The module '{mod_filename}' is missing a docstring."
            raise NoDocstringError(msg)
        res["summary"] = docstring.lstrip().splitlines()[0]

    if want_version:
        res["version"] = check_version(version, mod_filename)

    return res


def check_version(version, filename):
    """
    Check whether a given version string match PEP 440, and do normalisation.

    Raise InvalidVersion/NoVersionError with relevant information if
    version is invalid.

    Log a warning if the version is not canonical with respect to PEP 440.

    Returns the version in canonical PEP 440 format.
    """
    if not version:
        msg = f"Please define a `__version__ = \"x.y.z\"` in your module '{filename}'."
        raise NoVersionError(msg)
    if not isinstance(version, str):
        msg = f"__version__ must be a string, not {type(version)}, in module '{filename}'."
        raise InvalidVersion(msg)

    try:
        norm_version = distlib.version.NormalizedVersion(version)
        version = str(norm_version)
    except distlib.version.UnsupportedVersionError as e:
        msg = f"Invalid __version__ in module '{filename}'"
        raise InvalidVersion(msg) from e

    return version


# Own code
# --------------------------------------------------------------------------- #

from pyproject_metadata import StandardMetadata  # noqa: E402


def update_dynamic_metadata(metadata: StandardMetadata, mod_filename: Path | None):
    if mod_filename is None:
        if metadata.dynamic:
            msg = "If no module is specified, dynamic metadata is not allowed"
            raise ConfigError(msg)
        return
    res = get_info_from_module(mod_filename, metadata.dynamic)
    if "version" in res:
        metadata.version = res["version"]
    if "summary" in res:
        metadata.description = res["summary"]
    metadata.dynamic = []


def find_module(module_metadata: dict, src_dir: Path) -> Module:
    name: str = module_metadata["name"]
    base_dir: Path = src_dir / module_metadata["directory"]
    is_namespace: bool = module_metadata["namespace"]
    generated: str | None = module_metadata.get("generated")

    # If the module is to be generated later by CMake, don't search for it now
    if generated:
        return Module(
            name=name,
            # This won't have the right extension, but that's not an issue
            full_path=base_dir / name,
            base_path=src_dir,
            is_package=generated == "package",
            is_namespace=is_namespace,
            is_generated=True,
        )

    # Look for the module
    dir = lambda p: p.is_dir()
    file = lambda p: not is_namespace and p.is_file()
    options = [
        (base_dir / name, dir),
        (base_dir / "src" / name, dir),
        (base_dir / (name + ".py"), file),
        (base_dir / "src" / (name + ".py"), file),
    ]

    def check(p: Path, checker):
        return checker(p)

    found = list(filter(lambda x: check(*x), options))

    if len(found) > 1:
        msg = f"Module is ambiguous {name}: {', '.join(map(repr, sorted(found)))}"
        raise ConfigError(msg)
    elif not found:
        msg = f"No file/folder found for module {name}"
        raise ConfigError(msg)

    full_path: Path = found[0][0]
    is_package = found[0][1] == dir

    if is_package:
        init_exists = (full_path / "__init__.py").exists()
        if is_namespace and init_exists:
            msg = (
                "Namespace packages should not contain __init__.py " f"(in {full_path})"
            )
            raise ConfigError(msg)
        elif not is_namespace and not init_exists:
            msg = (
                f"Missing __init__.py in {full_path}. "
                "Perhaps you forgot to set "
                "tool.py-build-cmake.module.namespace=true?"
            )
            raise ConfigError(msg)

    return Module(
        name=name,
        full_path=full_path,
        base_path=src_dir,
        is_package=is_package,
        is_namespace=is_namespace,
    )



from __future__ import annotations

import re
from copy import copy
from typing import Any

from .native_tags import WheelTags, get_native_tags


def get_cross_tags(crosscfg: dict[str, Any]) -> WheelTags:
    """Get the PEP 425 tags to use when cross-compiling."""
    tags = get_native_tags()
    if "implementation" in crosscfg and "version" in crosscfg:
        tags["pyver"] = [crosscfg["implementation"] + crosscfg["version"]]
    if "abi" in crosscfg:
        tags["abi"] = [crosscfg["abi"]]
    if "arch" in crosscfg:
        tags["arch"] = [crosscfg["arch"]]
    return tags


def convert_abi_tag(abi_tag: str, cmake_cfg: dict | None) -> str:
    """Set the ABI tag to 'none' or 'abi3', depending on the config options
    specified by the user."""
    if not cmake_cfg:
        return "none"
    elif cmake_cfg["python_abi"] == "auto":
        return abi_tag
    elif cmake_cfg["python_abi"] == "none":
        return "none"
    elif cmake_cfg["python_abi"] == "abi3":
        # Only use abi3 if we're actually building for CPython
        m = re.match(r"^cp(\d+).*$", abi_tag)
        if m and int(m[1]) >= cmake_cfg["abi3_minimum_cpython_version"]:
            return "abi3"
        return abi_tag
    else:
        msg = "Unsupported python_abi"
        raise AssertionError(msg)


def convert_wheel_tags(tags: dict[str, list[str]], cmake_cfg: dict | None) -> WheelTags:
    """Apply convert_abi_tag to each of the abi tags."""
    tags = copy(tags)
    cvt_abi = lambda tag: convert_abi_tag(tag, cmake_cfg)
    tags["abi"] = list(map(cvt_abi, tags["abi"]))
    if "none" in tags["abi"]:
        tags["pyver"] = ["py3"]
    return tags


def is_pure(cmake_cfg: dict | None) -> bool:
    """Check if the package is a pure-Python package without platform-
    specific binaries."""
    if not cmake_cfg:
        return True
    return cmake_cfg["pure_python"]



from __future__ import annotations

import logging
from copy import copy, deepcopy
from dataclasses import dataclass

from ...common import ConfigError
from .config_path import ConfPath
from .dict import DictOfStrConfigOption
from .value_reference import ValueReference

logger = logging.getLogger(__name__)


@dataclass
class CMakeOption:
    values: dict[str, list[str | bool]]
    datatype: str | None = None
    strict: bool = True

    @classmethod
    def create(cls, x, datatype=None):
        assert datatype is None or isinstance(datatype, str)
        if isinstance(x, list):
            assert all(isinstance(e, (bool, str)) for e in x)
            return cls(values={"value": x}, datatype=datatype)
        assert isinstance(x, (bool, str))
        return cls(values={"value": [x]}, datatype=datatype)


class CMakeOptConfigOption(DictOfStrConfigOption):
    """
    Dictionary of CMake options, with optional data type, ability to merge
    or remove options, etc.
    """

    valid_keys = frozenset(
        ("+", "-", "=", "value", "append", "prepend", "type", "strict")
    )

    def get_typename(self, md: bool = False) -> str:
        return "dict (CMake)"

    @staticmethod
    def _convert_to_option(
        x: str | bool | list[str | bool] | dict[str, str | bool | list[str | bool]],
        pth: ConfPath,
    ):
        if isinstance(x, bool):
            return CMakeOption(values={"value": [x]}, datatype="BOOL")
        if isinstance(x, str):
            return CMakeOption(values={"value": [x]})
        if isinstance(x, list):
            all_bool = all(isinstance(b, bool) for b in x)
            nonempty = len(x) > 0
            return CMakeOption(
                values={"value": x},
                datatype=("BOOL" if all_bool and nonempty else None),
            )
        assert isinstance(x, dict)
        values: dict[str, list[str | bool]] = {}
        all_bool = True
        nonempty = False
        # Preserve lists and convert all values to singletons
        for k in ("+", "-", "=", "value", "append", "prepend"):
            if k not in x:
                continue
            v = x.pop(k)
            values[k] = v if isinstance(v, list) else [v]
            if not all(isinstance(b, bool) for b in values[k]):
                all_bool = False
            if len(values[k]) > 0:
                nonempty = True
        # Determine the data type
        datatype = x.pop("type", None)
        if datatype is None and all_bool and nonempty:
            datatype = "BOOL"
        if datatype is not None and not isinstance(datatype, str):
            msg = f'Error in {pth}: Type of "type" should be str, '
            msg += f"not {type(datatype)}"
            raise ConfigError(msg)
        # Check if the data type should be checked in strict mode or not
        strict = x.pop("strict", True)
        if not isinstance(strict, bool):
            msg = f'Error in {pth}: Type of "strict" should be bool, '
            msg += f"not {type(strict)}"
            raise ConfigError(msg)
        # Check if there are any unused keys left
        if x:
            msg = f"Error in {pth}: Invalid keys: {list(x)}"
            raise ConfigError(msg)
        return CMakeOption(values=values, datatype=datatype, strict=strict)

    def _combine_values(self, a: dict, b: dict):
        if "=" in b:
            a["value"] = deepcopy(b["="])
            return a
        if "value" in b:
            a["value"] = deepcopy(b["value"])
            return a
        a.setdefault("value", [])
        if "-" in b:
            remove = set(b["-"])
            a["value"] = [v for v in a["value"] if v not in remove]
        if "+" in b:
            a["value"] += b["+"]
        if "append" in b:
            a["value"] += b["append"]
        if "prepend" in b:
            a["value"] = b["prepend"] + a["value"]
        return a

    @staticmethod
    def _check_type(a, a_path):
        valid_types = (None, "STRING", "BOOL", "PATH", "FILEPATH")
        if a not in valid_types:
            msg = f"Unknown type '{a}' in {a_path}"
            raise ConfigError(msg)

    def _combine_types(
        self,
        a: str | None,
        b: str | None,
        strict: bool,
        a_path: ConfPath,
        b_path: ConfPath,
    ):
        def report(msg):
            if not strict:
                logger.warning(msg)
                return b
            msg = f'{msg} (use "strict" = false to ignore)'
            raise ConfigError(msg)

        if b is None:
            return a
        if a is None:
            return b
        if a == b:
            return b
        msg = "Incompatible types when overriding or inheriting CMake settings"
        # Cannot override these by other types
        if a in ("BOOL", "STRING", "FILEPATH"):
            return report(f"{msg}: {a} ({a_path}) and {b} ({b_path})")
        # PATH can be made more specific to FILEPATH
        if a == "PATH" and b != "FILEPATH":
            return report(f"{msg}: {a} ({a_path}) and {b} ({b_path})")
        return b

    def _combine(self, a: CMakeOption, b: CMakeOption, a_path, b_path):
        if "value" in b.values or "=" in b.values:
            return b
        return CMakeOption(
            values=self._combine_values(a.values, b.values),
            datatype=self._combine_types(
                a.datatype, b.datatype, b.strict, a_path, b_path
            ),
        )

    def override(self, old_value: ValueReference, new_value: ValueReference):
        # Entries are of the following form
        # {
        #    "VAR_A": "value",
        #    "VAR_B": ["123", "456"],
        #    "VAR_C": false,
        #    "VAR_D": {
        #       "type": "STRING",
        #       "value": ["789", "xyz"],
        #    },
        #    "VAR_E": {
        #       "type": "FILEPATH",
        #       "value": "/usr/bin/bash",
        #    },
        #    "VAR_F": {
        #       "type": "FILEPATH",
        #       "append": ["/usr/bin/bash"],
        #    }
        # }
        if old_value.values is None:
            old_value.values = {}
        if new_value.values is None:
            return old_value.values
        r = deepcopy(old_value.values)
        for k, v in new_value.values.items():
            if k not in r:
                r[k] = deepcopy(v)
            else:
                r[k] = self._combine(
                    a=r[k],
                    b=v,
                    a_path=old_value.value_path.join(k),
                    b_path=new_value.value_path.join(k),
                )
        return r

    def verify(self, values: ValueReference):
        """Checks the data types and keys of the values, and then converts
        them to a dictionary of CMakeOptions."""

        def verify_cmake_list(x, pth: ConfPath):
            if isinstance(x, list):
                for v in x:
                    if not isinstance(v, str) and not isinstance(v, bool):
                        msg = f"Type of values in list {pth} should be str or "
                        msg += f"bool, not {type(v)}"
                        raise ConfigError(msg)
            elif not isinstance(x, str) and not isinstance(x, bool):
                msg = f"Type of values in {pth} should be str or bool or a "
                msg += f"list thereof, not {type(x)}"
                raise ConfigError(msg)

        def verify_cmake_list_or_dict(x, pth: ConfPath):
            if isinstance(x, dict):
                for k, v in x.items():
                    if not isinstance(k, str):
                        msg = f"Type of keys in {pth} should be str, "
                        msg += f"not {type(k)}"
                        raise ConfigError(msg)
                    verify_cmake_list(v, pth.join(k))
            else:
                verify_cmake_list(x, pth)

        if values.values is None:
            return None
        valdict = copy(values.values)
        if not isinstance(valdict, dict):
            msg = f"Type of {values.value_path} should be {dict}, "
            msg += f"not {type(valdict)}"
            raise ConfigError(msg)
        elif not all(isinstance(k, str) for k in valdict):
            msg = f"Type of keys in {values.value_path} should be {str}"
            raise ConfigError(msg)
        for k, v in valdict.items():
            verify_cmake_list_or_dict(v, values.value_path.join(k))

        def check_dict_keys(d, pth):
            if not isinstance(d, dict):
                return
            invalid_keys = set(d) - self.valid_keys
            if invalid_keys:
                msg = f"Invalid keys in {pth}: {list(invalid_keys)}"
                raise ConfigError(msg)
            if "value" in d or "=" in d:
                invalid_keys = {"+", "-", "prepend", "append"} & set(d)
                if invalid_keys:
                    msg = 'Cannot combine "value" or "=" with the '
                    msg += f"following keys: {list(invalid_keys)}"
                    raise ConfigError(msg)
            if "type" in d:
                self._check_type(d["type"], pth)

        for k in valdict:
            pth = values.value_path.join(k)
            check_dict_keys(valdict[k], pth)
            valdict[k] = self._convert_to_option(valdict[k], pth)
        return valdict

    def finalize(self, values: ValueReference) -> dict[str, str]:
        """Converts the internal dict of CMakeOptions back into a dict of string
        that can be passed as CMake command-line arguments"""

        def convert(x):
            assert isinstance(x, CMakeOption)
            on_off = {True: "On", False: "Off"}
            escape = lambda x: x.replace(";", "\\;")
            to_str = lambda b: (escape(b) if isinstance(b, str) else on_off[b])
            return ";".join(map(to_str, x.values["value"]))

        def convert_key(k, x):
            assert isinstance(x, CMakeOption)
            return f"{k}:{x.datatype}" if x.datatype else k

        options: dict[str, str] = {}
        for k, v in values.values.items():
            options[convert_key(k, v)] = convert(v)
        return options



from .config_option import ConfigOption
from .config_path import ConfPath
from .config_reference import ConfigReference
from .default import (
    ConfigDefaulter,
    DefaultValueValue,
    MissingDefaultError,
    NoDefaultValue,
    RefDefaultValue,
    RequiredValue,
)
from .inherit import ConfigInheritor
from .override import ConfigOverrider
from .value_reference import ValueReference
from .verify import ConfigVerifier

__all__ = [
    "ConfigOption",
    "ConfPath",
    "ConfigReference",
    "ConfigDefaulter",
    "DefaultValueValue",
    "MissingDefaultError",
    "NoDefaultValue",
    "RefDefaultValue",
    "RequiredValue",
    "ConfigInheritor",
    "ConfigOverrider",
    "ValueReference",
    "ConfigVerifier",
]



from __future__ import annotations

from ...common import ConfigError
from .config_option import ConfigOption
from .config_path import ConfPath
from .default import DefaultValue, NoDefaultValue
from .value_reference import ValueReference


class EnumConfigOption(ConfigOption):
    def __init__(
        self,
        name: str,
        description: str = "",
        example: str = "",
        default: DefaultValue | None = None,
        inherit_from: ConfPath | str | None = None,
        create_if_inheritance_target_exists: bool = False,
        options: list[str] | None = None,
    ) -> None:
        if default is None:
            default = NoDefaultValue()
        if options is None:
            options = []
        super().__init__(
            name,
            description,
            example,
            default,
            inherit_from,
            create_if_inheritance_target_exists,
        )
        self.options = options

    def get_typename(self, md: bool = False):
        if md:
            return "`'" + "'` \\| `'".join(self.options) + "'`"
        else:
            return "'" + "' | '".join(self.options) + "'"

    def override(self, old_value: ValueReference, new_value: ValueReference):
        if new_value.values is None:
            return old_value.values
        return new_value.values

    def verify(self, values: ValueReference):
        if self.sub_options:
            msg = f"Type of {values.value_path} should be {str}, "
            msg += f"not {dict}"
            raise ConfigError(msg)
        elif not isinstance(values.values, str):
            msg = f"Type of {values.value_path} should be {str}, "
            msg += f"not {type(values.values)}"
            raise ConfigError(msg)
        if values.values not in self.options:
            msg = f"Value of {values.value_path} should be one of "
            msg += "'" + "', '".join(self.options) + "'"
            raise ConfigError(msg)
        return values.values



from __future__ import annotations

from .config_reference import ConfigReference
from .value_reference import ValueReference


class ConfigOverrider:
    def __init__(
        self,
        root: ConfigReference,
        ref: ConfigReference,
        values: ValueReference,
        new_values: ValueReference,
    ) -> None:
        self.ref = ref
        self.root = root
        self.values = values
        self.new_values = new_values

    def override(self):
        overridden_values = ValueReference(
            self.values.value_path,
            self.ref.config.override(self.values, self.new_values),
        )
        for name in self.ref.sub_options:
            ref = self.ref.sub_ref(name).resolve_inheritance(self.root)
            try:
                new_val = self.new_values.sub_ref(name)
            except KeyError:
                continue
            default = {} if ref.sub_options else None
            overridden_values.set_value_default(name, default)
            old_val = overridden_values.sub_ref(name)
            overridden_values.values[name] = ConfigOverrider(
                root=self.root,
                ref=ref,
                values=old_val,
                new_values=new_val,
            ).override()
        return overridden_values.values



from __future__ import annotations

from .config_path import ConfPath
from .config_reference import ConfigReference
from .override import ConfigOverrider
from .value_reference import ValueReference


class ConfigInheritor:
    def __init__(
        self,
        root: ConfigReference,
        root_values: ValueReference,
        ref: ConfigReference | None = None,
        value_path: ConfPath | None = None,
        done: set[tuple[str, ...]] | None = None,
    ) -> None:
        if ref is None:
            ref = root
        if value_path is None:
            value_path = ConfPath()
        if done is None:
            done = set()
        self.root = root
        self.root_values = root_values
        self.ref = ref
        self.value_path = value_path
        self.done = done

    def _inherit_self(self):
        # Check if this option inherits from another option
        inherits = self.ref.config.inherits
        if inherits is None:
            return
        # Only carry out the inheritance once
        if self.value_path.pth in self.done:
            return
        self.done.add(self.value_path.pth)
        # Find the option we inherit from
        super = self.ref.resolve_inheritance_single(self.root)
        # If the option we're inheriting from does not have its value set,
        # don't do anything.
        try:
            super_value = self.root_values.sub_ref(super.config_path)
        except KeyError:
            return
        # If our option does not have its value set, create this value and
        # all its parents.
        if not self._create_parent_values():
            return

        # If our super option inherits from other options, carry out that
        # inheritance first.
        ConfigInheritor(
            root=self.root,
            root_values=self.root_values,
            ref=super,
            value_path=super.config_path,
            done=self.done,
        )._inherit_self()
        # Note that we only detect direct inheritance, inheriting from a
        # subtree that is part of an inherited tree is not supported (and
        # and unnecessary for the current schema)

        # Create a copy of the values of our super option and override them
        # with our own values
        new_values = self.root_values.sub_ref(self.value_path)
        assert new_values is not None
        real_config = super.resolve_inheritance(self.root)
        inherited_values = ConfigOverrider(
            root=self.root,
            ref=real_config,
            values=super_value,
            new_values=new_values,
        ).override()
        self.root_values.set_value(self.value_path, inherited_values)

    def _create_parent_values(self):
        """
        Loop over all parent options of path in the root and default-initialize
        their values to an empty dict if the option's
        create_if_inheritance_target_exists member is set to True.
        """
        sub, val = self.root, self.root_values
        for s in self.value_path.pth:
            sub = sub.sub_ref(s)
            try:
                val = val and val.sub_ref(s)
            except KeyError:
                if not sub.config.create_if_inheritance_target_exists:
                    return False
                val = None
        val = self.root_values
        for s in self.value_path.pth:
            val.values.setdefault(s, {})
            val = val.sub_ref(s)
            assert val is not None
        return True

    def inherit(self):
        self._inherit_self()
        for name in self.ref.sub_options:
            ref = self.ref.sub_ref(name)
            val_path = self.value_path.join(name)
            ConfigInheritor(
                root=self.root,
                root_values=self.root_values,
                ref=ref,
                value_path=val_path,
                done=self.done,
            ).inherit()



"""
distlib.wheel doesn't always return the correct tags, and packaging.tags
returns all tags supported by the interpreter, not the tags that should be used
for the generated wheels. Therefore the only option here is to write our own
(kind of hacky) functions based on packaging.tags.
"""

from __future__ import annotations

import sys
import sysconfig
from importlib.machinery import EXTENSION_SUFFIXES
from typing import Dict, List

from distlib.util import (  # type: ignore[import-untyped]
    get_platform as get_platform_dashes,
)

from ..common.util import platform_to_platform_tag

_INTERPRETER_SHORT_NAMES: dict[str, str] = {
    "python": "py",
    "cpython": "cp",
    "pypy": "pp",
    "ironpython": "ip",
    "jython": "jy",
}


def get_platform_tag() -> str:
    return platform_to_platform_tag(get_platform_dashes())


def get_interpreter_name() -> str:
    name = sys.implementation.name
    return _INTERPRETER_SHORT_NAMES.get(name) or name


def get_interpreter_version() -> str:
    return "".join(map(str, sys.version_info[:2]))


def get_cpython_interpreter() -> str:
    return f"cp{get_interpreter_version()}"


def get_cpython_abi() -> str:
    """
    Get the ABI string for CPython, e.g. cp37m.

    https://github.com/pypa/packaging/blob/917612f5774571a99902b5fe04d06099b9e8b667/packaging/tags.py#L135
    """
    py_version = sys.version_info[:2]
    debug = pymalloc = ""
    with_debug = sysconfig.get_config_var("Py_DEBUG")
    has_refcount = hasattr(sys, "gettotalrefcount")
    # Windows doesn't set Py_DEBUG, so checking for support of debug-compiled
    # extension modules is the best option.
    # https://github.com/pypa/pip/issues/3383#issuecomment-173267692
    has_ext = "_d.pyd" in EXTENSION_SUFFIXES
    if with_debug or (with_debug is None and (has_refcount or has_ext)):
        debug = "d"
    if py_version < (3, 8):
        with_pymalloc = sysconfig.get_config_var("WITH_PYMALLOC")
        if with_pymalloc or with_pymalloc is None:
            pymalloc = "m"
    return f"{get_cpython_interpreter()}{debug}{pymalloc}"


def get_generic_interpreter() -> str:
    return f"{get_interpreter_name()}{get_interpreter_version()}"


def get_generic_abi() -> str:
    abi = sysconfig.get_config_var("SOABI") or "none"
    return platform_to_platform_tag(abi)


def get_python_tag() -> str:
    if get_interpreter_name() == "cp":
        return get_cpython_interpreter()
    else:
        return get_generic_interpreter()


def get_abi_tag() -> str:
    if get_interpreter_name() == "cp":
        return get_cpython_abi()
    else:
        return get_generic_abi()


WheelTags = Dict[str, List[str]]


def get_native_tags() -> WheelTags:
    """Get the PEP 425 tags for the current platform."""
    return {
        "pyver": [get_python_tag()],
        "abi": [get_abi_tag()],
        "arch": [get_platform_tag()],
    }



from __future__ import annotations

import contextlib
import logging
import os
from copy import copy
from dataclasses import dataclass, field
from pathlib import Path, PurePosixPath
from pprint import pprint
from typing import Any, Dict, Optional, cast

import pyproject_metadata
from distlib.util import normalize_name  # type: ignore[import-untyped]

from .. import __version__
from ..common import Config, ConfigError
from .options.config_path import ConfPath
from .options.config_reference import ConfigReference
from .options.default import ConfigDefaulter
from .options.finalize import ConfigFinalizer
from .options.inherit import ConfigInheritor
from .options.override import ConfigOverrider
from .options.pyproject_options import (
    get_component_options,
    get_cross_path,
    get_options,
    get_tool_pbc_path,
)
from .options.value_reference import ValueReference
from .options.verify import ConfigVerifier
from .quirks import config_quirks

try:
    import tomllib as toml_  # type: ignore[import,unused-ignore]
except ImportError:
    import tomli as toml_  # type: ignore[import,no-redef,unused-ignore]

logger = logging.getLogger(__name__)


def read_full_config(
    pyproject_path: Path,
    config_settings: dict[str, str | list[str]] | None,
    verbose: bool,
) -> Config:
    config_settings = config_settings or {}
    overrides = parse_config_settings_overrides(config_settings, verbose)
    cfg = read_config(pyproject_path, overrides)
    if verbose:
        print_config_verbose(cfg)
    return cfg


def parse_config_settings_overrides(
    config_settings: dict[str, str | list[str]], verbose: bool
):
    if verbose:
        print("Configuration settings:")
        pprint(config_settings)

    def listify(x: str | list[str]):
        return x if isinstance(x, list) else [x]

    def get_as_list(key: str):
        return listify(config_settings.get(key) or [])

    keys = ["local", "cross"]
    overrides = {key: get_as_list("--" + key) + get_as_list(key) for key in keys}
    if verbose:
        print("Configuration settings for local and cross-compilation overrides:")
        pprint(overrides)
    return overrides


def try_load_toml(path: Path):
    try:
        return toml_.loads(path.read_text("utf-8"))
    except FileNotFoundError as e:
        msg = f"Config file {str(path.absolute())!r} not found"
        raise ConfigError(msg) from e
    except OSError as e:
        msg = f"Config file {str(path.absolute())!r} could not be loaded"
        raise ConfigError(msg) from e
    except toml_.TOMLDecodeError as e:
        msg = f"Config file {str(path.absolute())!r} is invalid"
        raise ConfigError(msg) from e


def read_config(
    pyproject_path: str | Path, flag_overrides: dict[str, list[str]]
) -> Config:
    # Load the pyproject.toml file
    pyproject_path = Path(pyproject_path)
    pyproject_folder = pyproject_path.parent
    pyproject: dict[str, Any] = try_load_toml(pyproject_path)
    if "project" not in pyproject:
        msg = "Missing [project] table"
        raise ConfigError(msg)

    # Load local override
    localconfig_path = pyproject_folder / "py-build-cmake.local.toml"
    if localconfig_path.exists():
        flag_overrides.setdefault("local", []).insert(0, str(localconfig_path))
    # Load override for cross-compilation
    crossconfig_path = pyproject_folder / "py-build-cmake.cross.toml"
    if crossconfig_path.exists():
        flag_overrides.setdefault("cross", []).insert(0, str(crossconfig_path))

    # File names mapping to the actual dict with the config
    config_files: dict[str, dict[str, Any]] = {
        "pyproject.toml": pyproject,
        "<command-line>": {},  # FIXME: implement this
    }

    # Additional options for config_options
    overrides: dict[ConfPath, ConfPath] = {}
    # What to override
    extra_flag_paths = {"local": get_tool_pbc_path(), "cross": get_cross_path()}

    for flag, targetpath in extra_flag_paths.items():
        for path in map(Path, flag_overrides[flag]):
            if path.is_absolute():
                fullpath = path
            else:
                fullpath = (Path(os.environ.get("PWD", ".")) / path).resolve()
            config = try_load_toml(fullpath)
            if config:  # Treat empty file as no override
                config_files[fullpath.as_posix()] = config
                overrides[ConfPath((fullpath.as_posix(),))] = targetpath

    return process_config(pyproject_path, config_files, overrides)


def process_config(
    pyproject_path: Path | PurePosixPath,
    config_files: dict[str, dict[str, Any]],
    overrides: dict[ConfPath, ConfPath],
    test: bool = False,
) -> Config:
    pyproject = config_files["pyproject.toml"]
    assert pyproject is not None
    # Check the package/module name and normalize it
    f = "name"
    if f in pyproject["project"]:
        oldname = pyproject["project"][f]
        normname = normalize_name(oldname)
        if oldname != normname:
            logger.info("Name normalized from %s to %s", oldname, normname)
        pyproject["project"][f] = normname

    # Parse the [project] section for metadata
    try:
        Metadata = pyproject_metadata.StandardMetadata
        meta = Metadata.from_pyproject(pyproject, pyproject_path.parent)
    except pyproject_metadata.ConfigurationError as e:
        raise ConfigError(str(e)) from e

    # Create our own config data structure
    cfg = Config(meta)
    cfg.package_name = meta.name

    # py-build-cmake option tree
    opts = get_options(pyproject_path.parent, test=test)
    root_ref = ConfigReference(ConfPath.from_string("/"), opts)
    root_val = ValueReference(ConfPath.from_string("/"), config_files)

    # Verify the configuration and apply the overrides
    verify_and_override_config(overrides, root_ref, root_val)

    # Tweak the configuration depending on the environment and platform
    config_quirks(root_val.sub_ref(get_tool_pbc_path()))
    set_up_os_specific_cross_inheritance(root_ref, root_val)

    # Carry out inheritance between options
    inherit_default_and_finalize_config(root_ref, root_val)
    pbc_value_ref = root_val.sub_ref(get_tool_pbc_path())

    # Store the module configuration
    s = "module"
    if pbc_value_ref.is_value_set(s):
        # Normalize the import and wheel name of the package
        name_path = ConfPath((s, "name"))
        name = cast(str, pbc_value_ref.get_value(name_path))
        normname = name.replace("-", "_")
        pbc_value_ref.set_value(name_path, normname)
        cfg.module = pbc_value_ref.get_value(s)
    else:
        msg = "Missing [tools.py-build-cmake.module] section"
        raise AssertionError(msg)

    # Store the editable configuration
    cfg.editable = {
        os: cast(Dict[str, Any], pbc_value_ref.get_value(ConfPath((os, "editable"))))
        for os in ("linux", "windows", "mac", "cross")
        if pbc_value_ref.is_value_set(ConfPath((os, "editable")))
    }

    # Store the sdist folders (this is based on flit)
    def get_sdist_cludes(v: ValueReference) -> dict[str, Any]:
        return {
            clude + "_patterns": v.get_value(ConfPath(("sdist", clude)))
            for clude in ("include", "exclude")
        }

    cfg.sdist = {
        os: get_sdist_cludes(pbc_value_ref.sub_ref(os))
        for os in ("linux", "windows", "mac", "cross")
        if pbc_value_ref.is_value_set(os)
    }

    # Store the CMake configuration
    cfg.cmake = {
        os: cast(Dict[str, Any], pbc_value_ref.get_value(ConfPath((os, "cmake"))))
        for os in ("linux", "windows", "mac", "cross")
        if pbc_value_ref.is_value_set(ConfPath((os, "cmake")))
    }
    cfg.cmake = cfg.cmake or None

    # Store stubgen configuration
    s = "stubgen"
    if pbc_value_ref.is_value_set(s):
        cfg.stubgen = pbc_value_ref.get_value(s)

    # Store the cross compilation configuration
    s = "cross"
    if pbc_value_ref.is_value_set(s):
        cfg.cross = copy(cast(Optional[Dict[str, Any]], pbc_value_ref.get_value(s)))
        if cfg.cross is not None:
            for k in ("cmake", "sdist", "editable"):
                cfg.cross.pop(k, None)

    # Check for incompatible options
    cfg.check()

    return cfg


def verify_and_override_config(
    overrides: dict[ConfPath, ConfPath],
    root_ref: ConfigReference,
    root_val: ValueReference,
):
    root_val.set_value(
        "pyproject.toml",
        ConfigVerifier(
            root=root_ref,
            ref=root_ref.sub_ref("pyproject.toml"),
            values=root_val.sub_ref("pyproject.toml"),
        ).verify(),
    )
    for override, target in overrides.items():
        root_val.set_value(
            override,
            ConfigVerifier(
                root=root_ref,
                ref=root_ref.sub_ref(target),
                values=root_val.sub_ref(override),
            ).verify(),
        )
        root_val.set_value_default(target, {})
        root_val.set_value(
            target,
            ConfigOverrider(
                root=root_ref,
                ref=root_ref.sub_ref(target),
                values=root_val.sub_ref(target),
                new_values=root_val.sub_ref(override),
            ).override(),
        )


def inherit_default_and_finalize_config(
    root_ref: ConfigReference, root_val: ValueReference
):
    ConfigInheritor(
        root=root_ref,
        root_values=root_val,
    ).inherit()
    ConfigDefaulter(
        root=root_ref,
        root_values=root_val,
        ref=root_ref.sub_ref("pyproject.toml"),
        value_path=ConfPath.from_string("pyproject.toml"),
    ).update_default()
    root_val.set_value(
        "pyproject.toml",
        ConfigFinalizer(
            root=root_ref,
            ref=root_ref.sub_ref("pyproject.toml"),
            values=root_val.sub_ref("pyproject.toml"),
        ).finalize(),
    )


def set_up_os_specific_cross_inheritance(
    root_ref: ConfigReference, root_val: ValueReference
):
    """Update the cross-compilation configuration to inherit from the
    corresponding OS-specific configuration."""
    cross_os = None
    with contextlib.suppress(KeyError):
        os_path = get_tool_pbc_path().join("cross").join("os")
        cross_os = cast(Optional[str], root_val.get_value(os_path))

    if cross_os is not None:
        for s in ("cmake", "sdist", "editable"):
            parent = get_tool_pbc_path().join(cross_os).join(s)
            child = get_cross_path().join(s)
            root_ref.sub_ref(child).config.inherits = parent


def print_config_verbose(cfg: Config):
    print("\npy-build-cmake (" + __version__ + ")")
    print("options")
    print("================================")
    print("package_name:")
    print(repr(cfg.package_name))
    print("module:")
    pprint(cfg.module)
    print("editable:")
    pprint(cfg.editable)
    print("sdist:")
    pprint(cfg.sdist)
    print("cmake:")
    pprint(cfg.cmake)
    print("cross:")
    pprint(cfg.cross)
    print("stubgen:")
    pprint(cfg.stubgen)
    print("================================\n")


@dataclass
class ComponentConfig:
    standard_metadata: pyproject_metadata.StandardMetadata
    package_name: str = field(default="")
    component: dict[str, Any] = field(default_factory=dict)


def read_full_component_config(
    pyproject_path: Path, config_settings: dict[str, list[Any]] | None, verbose: bool
) -> ComponentConfig:
    config_settings = config_settings or {}
    cfg = read_component_config(pyproject_path)
    if cfg.standard_metadata.dynamic:
        msg = "Dynamic metadata not supported for components."
        raise ConfigError(msg)
    if verbose:
        print_component_config_verbose(cfg)
    return cfg


def read_component_config(pyproject_path: Path) -> ComponentConfig:
    # Load the pyproject.toml file
    pyproject = try_load_toml(pyproject_path)
    if "project" not in pyproject:
        msg = "Missing [project] table"
        raise ConfigError(msg)

    # File names mapping to the actual dict with the config
    config_files: dict[str, dict[str, Any]] = {
        "pyproject.toml": pyproject,
    }
    return process_component_config(pyproject_path, pyproject, config_files)


def process_component_config(
    pyproject_path: Path,
    pyproject: dict[str, Any],
    config_files: dict[str, dict[str, Any]],
):
    # Check the package/module name and normalize it
    f = "name"
    if f in pyproject["project"]:
        oldname = pyproject["project"][f]
        normname = normalize_name(oldname)
        if oldname != normname:
            logger.info("Name normalized from %s to %s", oldname, normname)
        pyproject["project"][f] = normname

        # Parse the [project] section for metadata
    try:
        meta = pyproject_metadata.StandardMetadata.from_pyproject(
            pyproject, pyproject_path.parent
        )
    except pyproject_metadata.ConfigurationError as e:
        raise ConfigError(str(e)) from e

    # Create our own config data structure
    cfg = ComponentConfig(meta)
    cfg.package_name = meta.name

    # py-build-cmake option tree
    opts = get_component_options(pyproject_path.parent)
    root_ref = ConfigReference(ConfPath.from_string("/"), opts)
    root_val = ValueReference(ConfPath.from_string("/"), config_files)

    # Verify the configuration and apply the overrides
    root_val.set_value(
        "pyproject.toml",
        ConfigVerifier(
            root=root_ref,
            ref=root_ref.sub_ref("pyproject.toml"),
            values=root_val.sub_ref("pyproject.toml"),
        ).verify(),
    )
    # Carry out inheritance between options
    ConfigInheritor(
        root=root_ref,
        root_values=root_val,
    ).inherit()
    ConfigDefaulter(
        root=root_ref,
        root_values=root_val,
        ref=root_ref.sub_ref("pyproject.toml"),
        value_path=ConfPath.from_string("pyproject.toml"),
    ).update_default()
    root_val.set_value(
        "pyproject.toml",
        ConfigFinalizer(
            root=root_ref,
            ref=root_ref.sub_ref("pyproject.toml"),
            values=root_val.sub_ref("pyproject.toml"),
        ).finalize(),
    )
    pbc_value_ref = root_val.sub_ref(get_tool_pbc_path())

    # Store the component configuration
    cfg.component = pbc_value_ref.get_value("component")

    return cfg


def print_component_config_verbose(cfg: ComponentConfig):
    print("\npy-build-cmake (" + __version__ + ")")
    print("options")
    print("================================")
    print("package_name:")
    print(repr(cfg.package_name))
    print("component:")
    pprint(cfg.component)
    print("================================\n")



from __future__ import annotations

from pathlib import Path, PurePosixPath

from ...common import CMAKE_MINIMUM_REQUIRED
from .bool import BoolConfigOption
from .cmake_opt import CMakeOptConfigOption
from .config_option import ConfigOption, UncheckedConfigOption
from .config_path import ConfPath
from .default import DefaultValueValue, NoDefaultValue, RefDefaultValue, RequiredValue
from .dict import DictOfStrConfigOption
from .dir_pattern import DirPatternsConfigOption
from .enum import EnumConfigOption
from .int import IntConfigOption
from .list import ListOfStrConfigOption
from .path import PathConfigOption, RelativeToCurrentConfig, RelativeToProject
from .string import StringConfigOption


def get_tool_pbc_path():
    return ConfPath.from_string("pyproject.toml/tool/py-build-cmake")


def get_cross_path():
    return ConfPath.from_string("pyproject.toml/tool/py-build-cmake/cross")


def get_options(project_path: Path | PurePosixPath, *, test: bool = False):
    root = ConfigOption("root")
    pyproject = root.insert(UncheckedConfigOption("pyproject.toml"))
    project = pyproject.insert(UncheckedConfigOption("project"))
    project.insert(UncheckedConfigOption("name", default=RequiredValue()))
    name_pth = ConfPath.from_string("pyproject.toml/project/name")
    tool = pyproject.insert(
        UncheckedConfigOption("tool",
                              default=DefaultValueValue({}),
                              create_if_inheritance_target_exists=True,
        ))  # fmt: skip
    pbc = tool.insert(
        ConfigOption("py-build-cmake",
                     default=DefaultValueValue({}),
                     create_if_inheritance_target_exists=True,
        ))  # fmt: skip

    # [tool.py-build-cmake.module]
    module = pbc.insert(
        ConfigOption("module",
                     "Defines the import name of the module or package, and "
                     "the directory where it can be found.",
                     default=DefaultValueValue({}),
        ))  # fmt: skip
    module.insert_multiple([
        StringConfigOption("name",
                           "Import name in Python (can be different from the "
                           "name on PyPI, which is defined in the [project] "
                           "section).",
                        default=RefDefaultValue(name_pth)),
        PathConfigOption("directory",
                         "Directory containing the Python module/package.",
                         default=DefaultValueValue("."),
                         base_path=RelativeToProject(project_path),
                         must_exist=not test),
        BoolConfigOption("namespace",
                         "Set to true for PEP 420 namespace packages.",
                         default=DefaultValueValue(False)),
        EnumConfigOption("generated",
                         "Do not try to locate the main module in the source "
                         "directory, but assume that it is generated by CMake. "
                         "Dynamic metadata cannot be used when set.",
                         options=["module", "package"]),
    ])  # fmt: skip

    # [tool.py-build-cmake.editable]
    editable = pbc.insert(
        ConfigOption("editable",
                     "Defines how to perform an editable install (PEP 660). "
                     "See https://tttapa.github.io/py-build-cmake/"
                     "Editable-install.html for more information.",
                     default=DefaultValueValue({}),
        ))  # fmt: skip
    editable_pth = ConfPath.from_string("pyproject.toml/tool/py-build-cmake/editable")
    editable.insert_multiple([
        EnumConfigOption("mode",
                         "Mechanism to use for editable installations. "
                         "Either write a wrapper __init__.py file, install an "
                         "import hook, or install symlinks to the original "
                         "files.",
                         default=DefaultValueValue("symlink"),
                         options=["wrapper", "hook", "symlink"]),
        BoolConfigOption("build_hook",
                         "Automatically re-build any changed files and C "
                         "extension modules when the package is first "
                         "imported by Python. It is recommended to use a fast "
                         "generator like Ninja. Currently, the only "
                         "mode that supports build hooks is `symlink`.",
                         default=DefaultValueValue(False)),
    ])  # fmt: skip

    # [tool.py-build-cmake.sdist]
    sdist = pbc.insert(
        ConfigOption("sdist",
                     "Specifies the files that should be included in the "
                     "source distribution for this package.",
                     default=DefaultValueValue({}),
                     create_if_inheritance_target_exists=True,
        ))  # fmt: skip
    sdist_pth = ConfPath.from_string("pyproject.toml/tool/py-build-cmake/sdist")
    sdist.insert_multiple([
        DirPatternsConfigOption("include",
                                "Files and folders to include in the source "
                                "distribution. May include the '*' wildcard "
                                "(but not '**' for recursive patterns).",
                                default=DefaultValueValue([])),
        DirPatternsConfigOption("exclude",
                                "Files and folders to exclude from the source "
                                "distribution. May include the '*' wildcard "
                                "(but not '**' for recursive patterns).",
                                default=DefaultValueValue([])),
    ])  # fmt: skip

    # [tool.py-build-cmake.cmake]
    cmake = pbc.insert(
        ConfigOption("cmake",
                     "Defines how to build the project to package. If omitted, "
                     "py-build-cmake will produce a pure Python package.",
        ))  # fmt: skip
    cmake_pth = ConfPath.from_string("pyproject.toml/tool/py-build-cmake/cmake")
    cmake.insert_multiple([
        StringConfigOption("minimum_version",
                           "Minimum required CMake version. Used for policies "
                           "in the automatically generated CMake cache pre-"
                           "load files. If this version is not available in "
                           "the system PATH, it will be installed "
                           "automatically as a build dependency (using Pip).",
                           "minimum_version = \"3.18\"",
                           default=DefaultValueValue(CMAKE_MINIMUM_REQUIRED)),
        StringConfigOption("build_type",
                           "Build type passed to the configuration step, as "
                           "`-DCMAKE_BUILD_TYPE=<?>`.",
                           "build_type = \"RelWithDebInfo\""),
        ListOfStrConfigOption("config",
                              "Configuration type passed to the build and "
                              "install steps, as `--config <?>`. You can "
                              "specify either a single string, or a list of "
                              "strings. If a multi-config generator is used, "
                              "all configurations in this list will be "
                              "included in the package.",
                              "config = [\"Debug\", \"Release\"]",
                              default=RefDefaultValue(
                                  ConfPath.from_string("build_type"),
                                  relative=True,
                              ),
                              convert_str_to_singleton=True),
        StringConfigOption("preset",
                           "CMake preset to use for configuration. Passed as "
                           "`--preset <?>` during the configuration phase."),
        ListOfStrConfigOption('build_presets',
                              "CMake presets to use for building. Passed as "
                              "`--preset <?>` during the build phase, once "
                              "for each preset.",
                              default=None,
                              convert_str_to_singleton=True),
        StringConfigOption("generator",
                           "CMake generator to use, passed to the "
                           "configuration step, as "
                           "`-G <?>`. If Ninja is used, and if it is not "
                           "available in the system PATH, it will be installed "
                           "automatically as a build dependency.",
                           "generator = \"Ninja Multi-Config\""),
        PathConfigOption("source_path",
                         "Folder containing CMakeLists.txt.",
                         default=DefaultValueValue("."),
                         expected_contents=[] if test else ["CMakeLists.txt"],
                         base_path=RelativeToProject(project_path),
                         must_exist=not test),
        PathConfigOption("build_path",
                         "CMake build and cache folder. The placeholder "
                         "`{build_config}` can be used to insert the name of "
                         "the Python version and ABI, operating system, and "
                         "architecture. This ensures that separate build "
                         "directories are used for different host systems and "
                         "Python versions/implementations.",
                         default=DefaultValueValue(".py-build-cmake_cache/{build_config}"),
                         allow_abs=True,
                         base_path=RelativeToProject(project_path),
                         must_exist=False),
        CMakeOptConfigOption("options",
                             "Extra options passed to the configuration step, "
                             "as `-D<option>=<value>`.",
                             "options = {\"WITH_FEATURE_X\" = true}",
                             default=DefaultValueValue({})),
        ListOfStrConfigOption("args",
                              "Extra arguments passed to the configuration "
                              "step.",
                              "args = [\"--debug-find\", \"-Wdev\"]",
                              default=DefaultValueValue([]),
                              append_by_default=True),
        BoolConfigOption("find_python",
                         "Specify hints for CMake's FindPython module.",
                         "find_python = true",
                         default=DefaultValueValue(False)),
        BoolConfigOption("find_python3",
                         "Specify hints for CMake's FindPython3 module.",
                         "find_python3 = false",
                         default=DefaultValueValue(True)),
        ListOfStrConfigOption("build_args",
                              "Extra arguments passed to the build step.",
                              "build_args = [\"-j\", \"--target\", \"foo\"]",
                              default=DefaultValueValue([]),
                              append_by_default=True),
        ListOfStrConfigOption("build_tool_args",
                              "Extra arguments passed to the build tool in the "
                              "build step (e.g. to Make or Ninja).",
                              "build_tool_args = "
                              "[\"--verbose\", \"-d\", \"explain\"]",
                              default=DefaultValueValue([]),
                              append_by_default=True),
        ListOfStrConfigOption("install_args",
                              "Extra arguments passed to the install step.",
                              "install_args = [\"--strip\"]",
                              default=DefaultValueValue([]),
                              append_by_default=True),
        ListOfStrConfigOption("install_components",
                              "List of components to install, the install step "
                              "is executed once for each component, with the "
                              "option `--component <?>`.\n"
                              "Use an empty string to specify the default "
                              "component.",
                              default=DefaultValueValue([""])),
        DictOfStrConfigOption("env",
                              "Environment variables to set when running "
                              "CMake. Supports variable expansion using "
                              "`${VAR}` (but not `$VAR`).",
                              "env = { \"CMAKE_PREFIX_PATH\" "
                              "= \"${HOME}/.local\" }",
                              default=DefaultValueValue({})),
        BoolConfigOption("pure_python",
                         "Indicate that this package contains no platform-"
                         "specific binaries, only Python scripts and other "
                         "platform-agnostic files. It causes the Wheel tags "
                         "to be set to `py3-none-any`.",
                         "pure_python = true",
                         default=DefaultValueValue(False)),
        EnumConfigOption("python_abi",
                         "Override the default ABI tag for the Wheel package.\n"
                         "For packages with a Python extension module that "
                         "make use of the full Python C API, this option "
                         "should be set to `auto`.\n"
                         "If your package does not contain Python extension "
                         "modules (e.g. because it only includes executables "
                         "to run as a subprocess, or only shared library files "
                         "to be loaded using `ctypes`), you can set this to "
                         "`none`.\n"
                         "If your package only includes Python extension "
                         "modules that use the CPython stable ABI, set this "
                         "`abi3` (see also `abi3_minimum_cpython_version` "
                         "below).\n"
                         "For details about platform compatibility tags, see "
                         "the PyPA specification: https://packaging.python.org/"
                         "en/latest/specifications/platform-compatibility-tags",
                         "python_abi = 'none'",
                         default=DefaultValueValue("auto"),
                         options=["auto", "none", "abi3"]),
        IntConfigOption("abi3_minimum_cpython_version",
                        "If `python_abi` is set to `abi3`, only use the stable "
                        "CPython API for CPython version that are newer than "
                        "`abi3_minimum_version`. Useful for nanobind, which "
                        "supports the stable ABI for CPython 12 and later.\n"
                        "The Python version is encoded as a single integer, "
                        "consisting of the major and minor version numbers, "
                        "without a dot.",
                        "abi3_minimum_cpython_version = 312",
                        default=DefaultValueValue(32)),
    ])  # fmt: skip

    # [tool.py-build-cmake.stubgen]
    stubgen = pbc.insert(
        ConfigOption("stubgen",
                     "If specified, mypy's stubgen utility will be used to "
                     "generate typed stubs for the Python files in the "
                     "package.",
        ))  # fmt: skip
    stubgen.insert_multiple([
        ListOfStrConfigOption("packages",
                              "List of packages to generate stubs for, passed "
                              "to stubgen as -p <?>."),
        ListOfStrConfigOption("modules",
                              "List of modules to generate stubs for, passed "
                              "to stubgen as -m <?>."),
        ListOfStrConfigOption("files",
                              "List of files to generate stubs for, passed to "
                              "stubgen without any flags."),
        ListOfStrConfigOption("args",
                              "List of extra arguments passed to stubgen.",
                              default=DefaultValueValue([]),
                              append_by_default=True),
    ])  # fmt: skip

    # [tool.py-build-cmake.{linux,windows,mac}]
    for system in ["Linux", "Windows", "Mac"]:
        name = system.lower()
        opt = pbc.insert(
            ConfigOption(name,
                         f"Override options for {system}.",
                         create_if_inheritance_target_exists=True,
                         default=DefaultValueValue({}),
            ))  # fmt: skip
        opt.insert_multiple([
            ConfigOption("editable",
                         f"{system}-specific editable options.",
                         inherit_from=editable_pth,
                         create_if_inheritance_target_exists=True),
            ConfigOption("sdist",
                         f"{system}-specific sdist options.",
                         inherit_from=sdist_pth,
                         create_if_inheritance_target_exists=True),
            ConfigOption("cmake",
                         f"{system}-specific CMake options.",
                         inherit_from=cmake_pth,
                         create_if_inheritance_target_exists=True),
        ])  # fmt: skip

    # [tool.py-build-cmake.cross]
    cross = pbc.insert(
        ConfigOption("cross",
                     "Causes py-build-cmake to cross-compile the project. See "
                     "https://tttapa.github.io/py-build-cmake/"
                     "Cross-compilation.html for more information.",
        ))  # fmt: skip
    cross.insert_multiple([
        EnumConfigOption("os",
                         "Operating system configuration to inherit from.",
                         options=["linux", "mac", "windows"]),
        StringConfigOption("implementation",
                           "Identifier for the Python implementation.\n"
                           "For details about platform compatibility tags, see "
                           "the PyPA specification: https://packaging.python.org/"
                           "en/latest/specifications/platform-compatibility-tags",
                           "implementation = 'cp' # CPython",
                           default=NoDefaultValue("same as current interpreter")),
        StringConfigOption("version",
                           "Python version, major and minor, without dots.\n"
                           "For details about platform compatibility tags, see "
                           "the PyPA specification: https://packaging.python.org/"
                           "en/latest/specifications/platform-compatibility-tags",
                           "version = '310' # 3.10",
                           default=NoDefaultValue("same as current interpreter")),
        StringConfigOption("abi",
                           "Python ABI.\n"
                           "For details about platform compatibility tags, see "
                           "the PyPA specification: https://packaging.python.org/"
                           "en/latest/specifications/platform-compatibility-tags",
                           "abi = 'cp310'",
                           default=NoDefaultValue("same as current interpreter")),
        StringConfigOption("arch",
                           "Operating system and architecture (no dots or "
                           "dashes, only underscores, all lowercase).\n"
                           "For details about platform compatibility tags, see "
                           "the PyPA specification: https://packaging.python.org/"
                           "en/latest/specifications/platform-compatibility-tags",
                           "arch = 'linux_x86_64'",
                           default=NoDefaultValue("same as current interpreter")),
        PathConfigOption("prefix",
                         "Root path of the Python installation. "
                         "Used to set the `Python3_ROOT_DIR` CMake hint, "
                         "see https://cmake.org/cmake/help/latest/module/"
                         "FindPython3.html#hints.",
                         base_path=RelativeToCurrentConfig(project_path),
                         allow_abs=True,
                         is_folder=True,
                         must_exist=True,
                         default=None),
        PathConfigOption("library",
                         "Python library file (.so on Linux, .lib on Windows). "
                         "Used to set the `Python3_LIBRARY` CMake artifact, "
                         "see https://cmake.org/cmake/help/latest/module/"
                         "FindPython3.html#artifacts-specification.",
                         base_path=RelativeToCurrentConfig(project_path),
                         allow_abs=True,
                         is_folder=False,
                         must_exist=True,
                         default=None),
        PathConfigOption("include_dir",
                         "Python include directory (containing Python.h). "
                         "Used to set the `Python3_INCLUDE_DIR` CMake "
                         "artifact, "
                         "see https://cmake.org/cmake/help/latest/module/"
                         "FindPython3.html#artifacts-specification.",
                         base_path=RelativeToCurrentConfig(project_path),
                         allow_abs=True,
                         is_folder=True,
                         must_exist=True,
                         default=None),
        PathConfigOption("toolchain_file",
                         "CMake toolchain file to use. See "
                         "https://cmake.org/cmake/help/book/mastering-cmake"
                         "/chapter/Cross%20Compiling%20With%20CMake.html for "
                         "more information.",
                         default=None,
                         base_path=RelativeToCurrentConfig(project_path),
                         must_exist=not test,
                         allow_abs=True,
                         is_folder=False),
        ConfigOption("editable",
                     "Override editable options when cross-compiling.",
                     inherit_from=editable_pth,
                     create_if_inheritance_target_exists=True),
        ConfigOption("sdist",
                     "Override sdist options when cross-compiling.",
                     inherit_from=sdist_pth,
                     create_if_inheritance_target_exists=True),
        ConfigOption("cmake",
                     "Override CMake options when cross-compiling.",
                     inherit_from=cmake_pth,
                     create_if_inheritance_target_exists=True),
    ])  # fmt: skip

    return root


def get_component_options(project_path: Path, *, test: bool = False):
    root = ConfigOption("root")
    pyproject = root.insert(UncheckedConfigOption("pyproject.toml"))
    project = pyproject.insert(UncheckedConfigOption("project"))
    project.insert(UncheckedConfigOption("name", default=RequiredValue()))
    tool = pyproject.insert(
        UncheckedConfigOption("tool",
                              default=DefaultValueValue({}),
                              create_if_inheritance_target_exists=True,
        ))  # fmt: skip
    pbc = tool.insert(
        ConfigOption("py-build-cmake",
                     default=DefaultValueValue({}),
                     create_if_inheritance_target_exists=True,
        ))  # fmt: skip

    # [tool.py-build-cmake.component]
    component = pbc.insert(
        ConfigOption("component",
                     "Options for a separately packaged component.",
                     default=DefaultValueValue({}),
        ))  # fmt: skip
    component.insert_multiple([
        PathConfigOption("main_project",
                         "Directory containing the main pyproject.toml file.",
                         default=DefaultValueValue(".."),
                         base_path=RelativeToProject(project_path),
                         must_exist=not test),
        ListOfStrConfigOption('build_presets',
                              "CMake presets to use for building. Passed as "
                              "`--preset <?>` during the build phase, once "
                              "for each preset.",
                              default=None,
                              convert_str_to_singleton=True),
        ListOfStrConfigOption("build_args",
                              "Extra arguments passed to the build step.",
                              "build_args = [\"-j\", \"--target\", \"foo\"]",
                              default=None,
                              append_by_default=True),
        ListOfStrConfigOption("build_tool_args",
                              "Extra arguments passed to the build tool in the "
                              "build step (e.g. to Make or Ninja).",
                              "build_tool_args = "
                              "[\"--verbose\", \"-d\", \"explain\"]",
                              default=None,
                              append_by_default=True),
        BoolConfigOption("install_only",
                         "Do not build the project, only install it.",
                         "install_only = true",
                         default=DefaultValueValue(False)),
        ListOfStrConfigOption("install_args",
                              "Extra arguments passed to the install step.",
                              "install_args = [\"--strip\"]",
                              default=None,
                              append_by_default=True),
        ListOfStrConfigOption("install_components",
                              "List of components to install, the install step "
                              "is executed once for each component, with the "
                              "option `--component <?>`.",
                              default=RequiredValue()),
    ])  # fmt: skip

    return root
